/* YAWN, Yet Another Web Navigator */
/* Copyright (C) 1997, 1998, James Troup <jjtroup@comp.brad.ac.uk>
 *                     Lee Willis <lwillis@comp.brad.ac.uk> */

/* This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include <gtk/gtk.h>
#include <string.h>
#include "neve.h"
#include "smg.h"
#include "menu.h"

/*****************************************************************************
 * function  : menu_init
 * arguments : broswer window
 * results   : 
 * purpose   : initializes a menu for tiffani
 * caveats   : 
 ***************************************************************************** */

GtkWidget *
menu_init (BrowserWindow * bw)
{
  GtkWidget *menubar;
  GtkWidget *file_menu;

  menubar = gtk_menu_bar_new ();

  bw->accelerator_table = gtk_accelerator_table_new ();
  gtk_accelerator_table_ref (bw->accelerator_table);

  file_menu = paula_menu_new (menubar, "File", bw);
  paula_menu_add (file_menu, "Quit", "<alt>Q", bw, menu_file_quit_callback);

  return menubar;
}

/*****************************************************************************
 * function  : paula_menu_new
 * arguments : menubar widget, name and broswer window
 * results   : 
 * purpose   : create a new root menu
 * caveats   : 
 ***************************************************************************** */

GtkWidget *
paula_menu_new (GtkWidget * menubar, const char *name, BrowserWindow * bw)
{
  GtkWidget *menu;
  GtkWidget *menuitem;

  menu = gtk_menu_new ();
  menuitem = gtk_menu_item_new_with_label ((const char *) name);
  gtk_menu_item_set_submenu (GTK_MENU_ITEM (menuitem), menu);
  gtk_menu_bar_append (GTK_MENU_BAR (menubar), menuitem);
  gtk_widget_show (menuitem);

  gtk_menu_set_accelerator_table (GTK_MENU (menu), bw->accelerator_table);

  return menu;
}

/*****************************************************************************
 * function  : paula_menu_add
 * arguments : menubar widget, name and broswer window
 * results   : 
 * purpose   : add a node to a root menu
 * caveats   : This is crude and won't handle dynamic menus in any way,
 *             shape or form
 ***************************************************************************** */

void
paula_menu_add (GtkWidget * menu, const char *name, const char *accelerator,
		BrowserWindow * bw, void (*callback) (GtkWidget * widget))
{
  GtkWidget *menuitem;
  char accelerator_key;
  guint8 accelerator_mods;

  menuitem = gtk_menu_item_new_with_label ((const char *) name);
  gtk_menu_append (GTK_MENU (menu), menuitem);
  gtk_widget_show (menuitem);
  if (accelerator != NULL)
    {
      gtk_menu_factory_parse_accelerator ((const char *) accelerator, &accelerator_key, &accelerator_mods);
      gtk_widget_install_accelerator (menuitem, bw->accelerator_table, "activate",
				      accelerator_key, accelerator_mods);
    }

  if (callback != NULL)		/* Why on earth would we have a dead menu item? */
    gtk_signal_connect (GTK_OBJECT (menuitem), "activate", (GtkSignalFunc) callback, NULL);
}

void
menu_file_quit_callback (GtkWidget * widget)
{
  gtk_main_quit ();
}

/* *****************************************************************************
 * NOTE: Stolen from GTK (gtk/gtkmenufactory.c) [and thus falls under the LGPL] 
 * ***************************************************************************** */

void
gtk_menu_factory_parse_accelerator (const char *accelerator,
				    char *accelerator_key,
				    guint8 * accelerator_mods)
{
  int done;

  *accelerator_key = 0;
  *accelerator_mods = 0;

  done = FALSE;
  while (!done)
    {
      if (strncmp (accelerator, "<shift>", 7) == 0)
	{
	  accelerator += 7;
	  *accelerator_mods |= GDK_SHIFT_MASK;
	}
      else if (strncmp (accelerator, "<alt>", 5) == 0)
	{
	  accelerator += 5;
	  *accelerator_mods |= GDK_MOD1_MASK;
	}
      else if (strncmp (accelerator, "<control>", 9) == 0)
	{
	  accelerator += 9;
	  *accelerator_mods |= GDK_CONTROL_MASK;
	}
      else
	{
	  done = TRUE;
	  *accelerator_key = accelerator[0];
	}
    }
}
