/* YAWN, Yet Another Web Navigator */
/* Copyright (C) 1998, James Troup <jjtroup@comp.brad.ac.uk>
 *                     Lee Willis <lwillis@comp.brad.ac.uk> */

/* This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include <stdio.h>
#include <stdlib.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <netdb.h>
#include <unistd.h>
#include "utils.h"
#include "globals.h" /* TEMP */
#include "neve.h"
#include "smg.h"
#include "config.h"
#include "header.h"
#include "chloe.h"
#include "http.h"

int make_socket (const char *hostname, const unsigned short int port);
void write_to_server (const int filedes, const char *message);
int read_from_client (const int filedes, char *buffer, const int bytes_to_read);
     
char *
get_via_http (const BrowserWindow *paula)
{

  int i, socket = 0;
  char *request, *header, *body;
  int header_length, body_length, body_counter;

  debug (debug_chloe, "chloe: connecting to %s%s:%d.", paula->hostname, paula->path, paula->port_number);
  socket = make_socket (paula->hostname, paula->port_number);
  debug (debug_chloe, "chloe: socket = %d", socket);
  request = xmalloc (40000);
  sprintf(request, "GET %s HTTP/1.0
User-Agent: %s/%s
Host: %s:%d
Accept: */*

", paula->path, PACKAGE, VERSION, paula->hostname, paula->port_number);
  debug (debug_chloe, "---request begin---\n%s---request end---", request);
  write_to_server (socket, request);
  debug (debug_chloe, "chloe: request sent.");
  header = xmalloc (1024);
  header_length =  read_from_client (socket, header, 1024);
  debug (debug_chloe, "Jeri: header_length is %d", header_length);
  {
    int foo = 0;
    debug (debug_chloe, "---begin header---");
    while (foo < 1024 && header[foo] != '\0')
      {
	if ((debug_chloe & debug_level))
	  fprintf (stderr, "%c", header[foo]);
	foo++;
      }
    debug (debug_chloe, "\n---end header---");
  }
  i =  parse_header (header, header_length, &body, &body_length, &body_counter);

  while (body_counter < body_length)
    {
      debug (debug_chloe, "chloe: at %d/%d in buffer, passing %p+%d => %p\n", body_counter, body_length,
	     body, body_counter, body_counter+body);
      i = read_from_client (socket, body+body_counter, body_length - body_counter);
      debug (debug_chloe, "chloe: read %d bytes", i);
      if (i >= 0)
	body_counter += i;
      else
	error ("AIEE");
   }
  /* Terminate the buffer */
  body[body_length] = '\0';

  debug (debug_chloe, "chloe: all done; outie.");

  return body;
}

/* Internal stuff */

int
make_socket (const char *hostname, const unsigned short int port)
{
  int sock;
  struct sockaddr_in name;
  struct hostent *hostinfo;
     
  /* Create the socket. */
  sock = socket (PF_INET, SOCK_STREAM, 0);
  if (sock < 0)
    {
      perror ("socket");
      exit (EXIT_FAILURE);
    }
     
  /* Give the socket a name. */
  name.sin_family = AF_INET;
  name.sin_port = htons (port);
  hostinfo = gethostbyname (hostname);
  if (hostinfo == NULL)
    {
      debug (debug_chloe, "Unknown host %s.", hostname);
      exit (EXIT_FAILURE);
    }
/*   memcpy (&name.sin_addr, hostinfo->h_addr_list[0], hostinfo->h_length); */
  name.sin_addr = *(struct in_addr *) hostinfo->h_addr;
  if (connect (sock, (struct sockaddr *) &name, sizeof (name)) < 0)
    {
      perror ("connect");
      exit (EXIT_FAILURE);
    }
     
  return sock;
}

void
write_to_server (const int filedes, const char *message)
{
  int nbytes;
     
  nbytes = write (filedes, message, strlen (message) + 1);
  if (nbytes < 0)
    {
      perror ("write");
      exit (EXIT_FAILURE);
    }
}

int
read_from_client (const int filedes, char *buffer, const int bytes_to_read)
{
  int nbytes;

  nbytes = read (filedes, buffer, bytes_to_read);
  if (nbytes < 0)
    {
      fubar (SYSERR, "read error in chloe");
      return -2;
    }
  else if (nbytes == 0)
    /* End-of-file. */
    return -1;
  else
    return nbytes;
}
     
