/* YAWN, Yet Another Web Navigator */
/* Copyright (C) 1998, James Troup <jjtroup@comp.brad.ac.uk>
 *                     Lee Willis <lwillis@comp.brad.ac.uk> */

/* This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

/* This code; like the rest of chloe, borrows a lot from wget */

#include <stdio.h>
#include <string.h>
#include <ctype.h>
#include "neve.h"
#include "smg.h"
#include "utils.h"
#include "url.h"

#define DEFAULT_HTTP_PORT 80

/* No.  This is not a joke.  See http://www.w3.org/Addressing/schemes.html */

#ifdef WE_DONT_USE_THIS_YET
static const char *schemes[] =
{
  "about", "callto", "cid", "clsid", "data", "fax", "file", "finger", "ftp", 
  "gopher", "hdl", "http", "https", "ilu", "imap", "IOR", "irc", "isbn", 
  "java", "javascript", "jdbc", "ldap", "lifn", "livescript", "mailto", 
  "mailserver", "md5", "mid", "mocha", "modem", "news", "nfs", "nntp", "path",
  "prospero", "res", "rvp", "rwhois", "rx", "service", "sip", "shhtp", "snews",
  "STANF", "telephone", "telnet", "tv", "uuid", "vemmi", "videotex", 
  "view-source", "wais", "whois++", "whodp", "z39.50r", "z39.50s", NULL
};
#endif 

#define IS_SCHEME_CHAR(c) (isalnum ((unsigned char)(c))	\
			   || (c) == '+' || (c) == '.' || (c) == '-')

void parse_http_url (const char *url, BrowserWindow *paula);
void parse_file_url (const char *url, BrowserWindow *paula);

int
parse_url (const char *url, BrowserWindow *paula)
{

  char *p;

  for (p = url; IS_SCHEME_CHAR (*p); p++);
  if (*p == ':' && p != url)
    { /* Absolute URL */
      /* Determine the scheme. 
       *
       * Currently, we support two schemes:
       *
       *   (1) HTTP
       *   (2) Local files ("file://localhost/" or "file:/") */
      
      if (!strncasecmp (url, "http://", 7))
	parse_http_url (url+7, paula);
      else if (!(strncasecmp (url, "file://localhost/", 17)))
	parse_file_url (url+16, paula);
      else if (!(strncasecmp (url, "file:/", 6)))
	parse_file_url (url+5, paula);
      else
	{
	  error ("%s: Only HTTP and (local) file schemes are supported", url);
	  return -1;
	}
    }
  else
    fubar (NONSYS, "Game over man!  What are we going to do now?");

  return 0;

}

void
parse_http_url (const char *url, BrowserWindow *paula)
{

  unsigned int i = 0;
  unsigned int length_of_url = strlen (url);

  paula->scheme = http;

  paula->port_number = DEFAULT_HTTP_PORT;

  /* Determine the net location */

  while (i < length_of_url && url[i] != '/')
    i++;
  
  paula->hostname = xmalloc (i + 1);
  strncpy (paula->hostname, url, i);
  
  /* Determine the path */

  if (i < length_of_url)
    { /* There is a path */
      paula->path = xmalloc ((length_of_url - i) + 1);
      strncpy (paula->path, url+i, (length_of_url - i));
    }
  else
    { /* There is no path; use default */
      paula->path = xmalloc (2);
      strcpy (paula->path, "/");
    }

  debug (debug_chloe, "parse_url: %s -> http://%s:%d%s", 
	 url, paula->hostname, paula->port_number, paula->path);

}

void
parse_file_url (const char *url, BrowserWindow *paula)
{

  int i;

  paula->scheme = file;

  if (url[0] != '/')
    { /* Relative filename */
      error ("relative URL?  Surely shome mistake: %s", url);
      paula->hostname = xmalloc (1025);
      getcwd (paula->hostname, 1024);
      i = 0;
    }
  else
    { /* Absolute filename */
      i = strlen (url);
      while (url[i] != '/')
	i--;
      i++;
      paula->hostname = xmalloc (i+1);
      strncpy (paula->hostname, url, i);
    }      
  
  paula->path = xmalloc ((strlen (url) - i) + 1);
  strcpy (paula->path, url+i);
  debug (debug_chloe, "file://localhost/%s%s", paula->hostname, paula->path);
  
}

void
absolutize_url (char *url, BrowserWindow *paula)
{

  unsigned int i, new_url_length;
  int last_slash_position_in_url, last_slash_position_in_path;
  char *new_url = NULL;
  char *p;



  for (p = url; IS_SCHEME_CHAR (*p); p++);
  if (*p == ':' && p != url)
    debug (debug_chloe, "absolutize_url: skipping absolute URL:%s", url);
  else
    { 
      debug (debug_chloe, "absolutize_url: found relative URL:%s", url);

      last_slash_position_in_url = -1;

      /* Scheme */
      
      switch (paula->scheme)
	{
	case file:
	  new_url_length += 17;
	  break;
	case http:
	  new_url_length += 7;
	  break;
	default:
	  fubar (NONSYS, "absolutize_url: unreachable code reached in switch (paula->scheme)");
	}

      /* Hostname */
      
      new_url_length += strlen (paula->hostname);

      /* There are three alternatives for a relative URL:
       *
       *   (1) absolute path
       *   (2) relative path
       *   (3) filename only */
      
      /* Absolute path (1) */

      if (url[0] == '/')
	{
	  
	  new_url_length += strlen (url);
	  new_url_length ++; /* Null terminator */
	  new_url = xmalloc (new_url_length);

	  strcpy (new_url, (paula->scheme == file) ? "file://localhost/" : "http://");
	  strcat (new_url, paula->hostname);
	  strcat (new_url, url);
	  debug (debug_chloe, "absolute_url: (1) %s", new_url);
	}
      /* Relative path (2) and  Filename only (3) */
      else
	{
	  i = 0;
	  while (i < strlen (url))
	    {
	      if (url[i] == '/')
		last_slash_position_in_url = i;
	      i++;
	    }

	  if (last_slash_position_in_url == -1)
	    {
	      /* Filename only (3) */

	      i = 0;
	      last_slash_position_in_url = -1;
	      while (i < strlen (paula->path))
		{
		  if (paula->path[i] == '/')
		    last_slash_position_in_path = i;
		  i++;
		}
	      /* Sanity check */
	      if (last_slash_position_in_path == -1)
		fubar (NONSYS, "absolutize_url: non-sequitor; paula's pathname has no directory component.");

	      new_url_length += last_slash_position_in_path + 1;
	 
	      new_url_length += strlen (url);
	      new_url_length ++; /* Null terminator */
	      new_url = xmalloc (new_url_length);

	      strcpy (new_url, (paula->scheme == file) ? "file://localhost/" : "http://");
	      strcat (new_url, paula->hostname);
	      strncat (new_url, paula->path, last_slash_position_in_path+1);
	      strcat (new_url, url);
	      debug (debug_chloe, "absolutize_url: (3) %s", new_url);
	    }
	  else
	    {
	      /* Relative path (2) */
	      /* FIXME: we don't handle ".." etc. */

	      i = 0;
	      last_slash_position_in_url = -1;
	      while (i < strlen (paula->path))
		{
		  if (paula->path[i] == '/')
		    last_slash_position_in_path = i;
		  i++;
		}
	      /* Sanity check */
	      if (last_slash_position_in_path == -1)
		fubar (NONSYS, "absolutize_url: non-sequitor; paula's pathname has no directory component.");

	      new_url_length += last_slash_position_in_path + 1;
	 
	      new_url_length += strlen (url);
	      new_url_length ++; /* Null terminator */
	      new_url = xmalloc (new_url_length);

	      strcpy (new_url, (paula->scheme == file) ? "file://localhost/" : "http://");
	      strcat (new_url, paula->hostname);
	      strncat (new_url, paula->path, last_slash_position_in_path+1);
	      strcat (new_url, url);
	      debug (debug_chloe, "absolutize_url: (2) %s", new_url);
	    }
	}
    }

  if (new_url)
    strcpy (url, new_url);

}
