/* YAWN, Yet Another Web Navigator */
/* Copyright (C) 1998, James Troup <jjtroup@comp.brad.ac.uk>
 *                     Lee Willis <lwillis@comp.brad.ac.uk> */

/* This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include <stdio.h>
#include <string.h>
#include <gtk/gtkmain.h>
#include <gtk/gtksignal.h>
#include <gdk/gdkkeysyms.h>
#include "utils.h"
#include "neve.h"
#include "alyson.h"
#include "urlentry.h"
#include "lisa.h"
#include "lisa_scrolling.h"
#include "lisa_expose.h"
#include "url.h"
#include "chloe.h"
#include "parser.h"

/* Private declarations */

static void gtk_lisa_class_init (GtkLisaClass * klass);
static void gtk_lisa_init (GtkLisa * lisa);
static void gtk_lisa_destroy (GtkObject * object);
static void gtk_lisa_realize (GtkWidget * widget);
static void gtk_lisa_unrealize (GtkWidget * widget);
static void gtk_lisa_size_request (GtkWidget * widget,
				   GtkRequisition * requisition);
static void gtk_lisa_size_allocate (GtkWidget * widget,
				    GtkAllocation * allocation);
static gint gtk_lisa_button_press (GtkWidget * widget,
				   GdkEventButton * event);
static gint gtk_lisa_key_press (GtkWidget * widget,
				GdkEventKey * event);
void set_vertical_scroll (GtkLisa *lisa);
static void recompute_geometry (GtkLisa* lisa);

static GtkWidgetClass *parent_class = NULL;

/* A lot of this is GTK+ widget stuff, which is black box, black magic
   to me. */

guint
gtk_lisa_get_type ()
{
  static guint lisa_type = 0;

  /* \begin{rant}Why can't the fscking GTK+ people use const?  Okay
   * there are some lame compilers out there which can't handle it,
   * but this can be trivially worked around with a #define.  As it
   * is, no gtk_*() calls afaics use const, meaning gcc bitches at me
   * whenever I pass a const to it (e.g a "foo").\end{rant} */

  if (!lisa_type)
    {
      GtkTypeInfo lisa_info =
      {
	"GtkLisa",
	sizeof (GtkLisa),
	sizeof (GtkLisaClass),
	(GtkClassInitFunc) gtk_lisa_class_init,
	(GtkObjectInitFunc) gtk_lisa_init,
	(GtkArgGetFunc) NULL,
	(GtkArgSetFunc) NULL
      };
      
      lisa_type = gtk_type_unique (gtk_widget_get_type (), &lisa_info);
    }

  return lisa_type;
}

static void
gtk_lisa_class_init (GtkLisaClass * class)
{
  GtkObjectClass *object_class;
  GtkWidgetClass *widget_class;

  object_class = (GtkObjectClass *) class;
  widget_class = (GtkWidgetClass *) class;

  parent_class = gtk_type_class (gtk_widget_get_type ());

  object_class->destroy = gtk_lisa_destroy;

  widget_class->realize = gtk_lisa_realize;
  widget_class->unrealize = gtk_lisa_unrealize;
  widget_class->expose_event = gtk_lisa_expose;
  widget_class->size_request = gtk_lisa_size_request;
  widget_class->size_allocate = gtk_lisa_size_allocate;
  widget_class->button_press_event = gtk_lisa_button_press;
  widget_class->key_press_event = gtk_lisa_key_press;
}

static void
gtk_lisa_init (GtkLisa * lisa)
{

  lisa->hadj = NULL;
  lisa->vadj = NULL;
  lisa->last_ver_value = 0;
  lisa->button = 0;
  lisa->offset = 12;
  lisa->virtual_y_offset = 0;
  lisa->line_cache = NULL;
  lisa->neve = NULL;
  lisa->in_a_link = FALSE;
  lisa->background_colour_set = FALSE;
  lisa->link_colour = NULL;
  lisa->visited_link_colour = NULL;
  lisa->active_link_colour = NULL;

}

GtkWidget *
gtk_lisa_new (GtkAdjustment * hadj,
	      GtkAdjustment * vadj)
{
  GtkLisa *lisa;

  lisa = gtk_type_new (gtk_lisa_get_type ());

  gtk_lisa_set_adjustments (lisa, hadj, vadj);

  return GTK_WIDGET (lisa);
}

static void
gtk_lisa_destroy (GtkObject * object)
{
  GtkLisa *lisa;

  if (!object)
    return;
  if (!GTK_IS_LISA (object))
    return;

  lisa = GTK_LISA (object);
  if (lisa->hadj)
    {
      gtk_object_unref (GTK_OBJECT (lisa->hadj));
      lisa->hadj = NULL;
    }
  if (lisa->vadj)
    {
      gtk_object_unref (GTK_OBJECT (lisa->vadj));
      lisa->vadj = NULL;
    }
  if (lisa->link_colour)
    free (lisa->link_colour);
  if (lisa->visited_link_colour)
    free (lisa->visited_link_colour);
  if (lisa->active_link_colour)
    free (lisa->active_link_colour);
  
  /* FIXME: lots to free */

  if (GTK_OBJECT_CLASS (parent_class)->destroy)
    (*GTK_OBJECT_CLASS (parent_class)->destroy) (object);
}

static void
gtk_lisa_realize (GtkWidget * widget)
{
  GtkLisa *lisa;
  GdkWindowAttr attributes;
  gint attributes_mask;
  GdkColor bg_colour;
  GdkColormap *cmap;

  if (!widget)
    return;
  if (!GTK_IS_LISA (widget))
    return;

  lisa = GTK_LISA (widget);
  GTK_WIDGET_SET_FLAGS (lisa, GTK_REALIZED);

  attributes.x = widget->allocation.x;
  attributes.y = widget->allocation.y;
  attributes.width = widget->allocation.width;
  attributes.height = widget->allocation.height;
  attributes.wclass = GDK_INPUT_OUTPUT;
  attributes.window_type = GDK_WINDOW_CHILD;
  attributes.visual = gtk_widget_get_visual (widget);
  attributes.colormap = gtk_widget_get_colormap (widget);
  attributes.event_mask = gtk_widget_get_events (widget);
  attributes.event_mask |= (GDK_EXPOSURE_MASK |
			    GDK_BUTTON_PRESS_MASK |
			    GDK_ENTER_NOTIFY_MASK |
			    GDK_LEAVE_NOTIFY_MASK |
			    GDK_KEY_PRESS_MASK);
  attributes_mask = GDK_WA_X | GDK_WA_Y | GDK_WA_VISUAL | GDK_WA_COLORMAP;

  widget->window = gdk_window_new (gtk_widget_get_parent_window (widget), &attributes, attributes_mask);
  gdk_window_set_user_data (widget->window, lisa);

  attributes.x = (widget->style->klass->xthickness + 1);
  attributes.y = (widget->style->klass->ythickness + 1);
  attributes.width = widget->allocation.width - attributes.x * 2;
  attributes.height = widget->allocation.height - attributes.y * 2;

  lisa->nicole = gdk_window_new (widget->window, &attributes, attributes_mask);
  gdk_window_set_user_data (lisa->nicole, lisa);

  widget->style = gtk_style_attach (widget->style, widget->window);

  /* FIXME: This really, so isn't the right way to set the background */

  cmap = gtk_widget_get_colormap (&lisa->widget);
  bg_colour.red = 48400;
  bg_colour.green = 48400;
  bg_colour.blue = 48400;
  if (!gdk_color_alloc (attributes.colormap, &bg_colour))
    fubar (NONSYS, "gtk_lisa_realize: couldn't allocate colour");
  gdk_window_set_background (lisa->nicole, &bg_colour);
  gdk_window_set_background (widget->window, &bg_colour);

  lisa->gc = gdk_gc_new (lisa->nicole);
  gdk_gc_set_exposures (lisa->gc, TRUE);
  gdk_gc_set_foreground (lisa->gc, &widget->style->text[GTK_STATE_NORMAL]);

  gdk_window_show (lisa->nicole);

  if ((widget->allocation.width > 1) || (widget->allocation.height > 1))
    {
      debug (debug_lisa, "Recomputing gemoetery");
      recompute_geometry (lisa); 
    }
}

static void
gtk_lisa_unrealize (GtkWidget * widget)
{
  GtkLisa *lisa;

  if (!widget)
    return;
  if (!GTK_IS_LISA (widget))
    return;

  lisa = GTK_LISA (widget);

  gdk_window_set_user_data (lisa->nicole, NULL);
  gdk_window_destroy (lisa->nicole);
  lisa->nicole = NULL;

  gdk_gc_destroy (lisa->gc);
  lisa->gc = NULL;

  if (GTK_WIDGET_CLASS (parent_class)->unrealize)
    (*GTK_WIDGET_CLASS (parent_class)->unrealize) (widget);
}

static void
gtk_lisa_size_request (GtkWidget * widget,
		       GtkRequisition * requisition)
{

  if (!widget)
    return;
  if (!GTK_IS_LISA (widget))
    return;
  if (!requisition)
    return;

  requisition->width = 10;
  requisition->height = 10;
}

void
set_vertical_scroll (GtkLisa *lisa)
{
  gint height;

  gdk_window_get_size (lisa->nicole, NULL, &height);

  lisa->vadj->step_increment = MIN (lisa->vadj->upper, (float) SCROLL_PIXELS);
  lisa->vadj->page_increment = MIN (lisa->vadj->upper, height);
  lisa->vadj->page_size      = MIN (lisa->vadj->upper, height);
  lisa->vadj->value          = MIN (lisa->vadj->value, lisa->vadj->upper - lisa->vadj->page_size);
  lisa->vadj->value          = MAX (lisa->vadj->value, 0.0);

  lisa->last_ver_value = (gint)lisa->vadj->value;

  gtk_signal_emit_by_name (GTK_OBJECT (lisa->vadj), "changed");

}

static void
recompute_geometry (GtkLisa* lisa)
{

  /*  free_louise (); */

  if (lisa->neve)
    {
      GdkRectangle render_area;
      gint width, height;

      neve_to_louise (lisa->neve, lisa);

      gdk_window_get_size (lisa->nicole, &width, &height);
      render_area.x = 0;
      render_area.y = 0;
      render_area.width = width;
      render_area.height = height;
      process_expose (lisa, &render_area, 0);

    }
  set_vertical_scroll (lisa);

}

static void
gtk_lisa_size_allocate (GtkWidget * widget,
			GtkAllocation * allocation)
{
  GtkLisa *lisa;

  debug (debug_lisa, "lisa_size_allocate: starting");

  if (!widget)
    return;
  if (!GTK_IS_LISA (widget))
    return;
  if (!allocation)
    return;

  widget->allocation = *allocation;
  if (GTK_WIDGET_REALIZED (widget))
    {
      lisa = GTK_LISA (widget);

      gdk_window_move_resize (widget->window,
			      allocation->x, allocation->y,
			      allocation->width, allocation->height);

      gdk_window_move_resize (lisa->nicole,
			      widget->style->klass->xthickness + 1,
			      widget->style->klass->ythickness + 1,
			      widget->allocation.width - (widget->style->klass->xthickness +
							  1) * 2,
			      widget->allocation.height - (widget->style->klass->ythickness +
							   1) * 2);
      debug (debug_lisa, "lisa_size_allocate: recomputing geometry (new -> %dx%d)", 
	     allocation->width, allocation->height);
      recompute_geometry (lisa);

    }
}

static gint
gtk_lisa_button_press (GtkWidget * widget,
		       GdkEventButton * event)
{
  GtkLisa *lisa;
  charisma *links;

  if (!widget)
    return FALSE;
  if (!GTK_IS_LISA (widget))
    return FALSE;
  if (!event)
    return FALSE;

  debug (debug_lisa, "gtk_lisa_button_press()");

  lisa = GTK_LISA (widget);
  links = lisa->links;


  /* FIXME: This is for tracking mouse clicks, but doesn't work till
   * we implement release_button.  I'm not entirely sure we need to
   * know about button releases.  I think gtktext uses it for
   * selection... */
#ifdef 0
  if (lisa->button && (event->button != lisa->button))
    return FALSE;
  lisa->button = event->button;
#endif /* 0 */
  /* At some stage being able to select text to copy into X's buffers
     would be nice, but uh, not any time soon */

  if (!GTK_WIDGET_HAS_FOCUS (widget))
    gtk_widget_grab_focus (widget);

  if (event->type == GDK_BUTTON_PRESS && event->button == 1)
    {
      int real_y = (int) event->y + lisa->virtual_y_offset;
      
      if (!links)
	{
	  debug (debug_lisa, "gtk_lisa_button_press: no links");
	  return FALSE;
	}

      while (links)
	{
	  if (event->x >= links->area->x &&
	      event->x <= links->area->x + links->area->width &&
	      real_y >= links->area->y &&
	      real_y <= links->area->y + links->area->height)
	    {
	      debug (debug_lisa, "SELECTED with %fx%d in \"%s\" at %dx%d.%d+%d (%d)",
		     event->x, real_y, (links->href) ? links->href : "",
		     links->area->x, links->area->y,
		     links->area->width, links->area->height, lisa->virtual_y_offset);
	      if (!links->href)
		gdk_beep ();
	      else
		{
		  BrowserWindow *paula;
		  char *body;
		  int i;

		  paula = help_me_bob();
		  urlentry_update (paula->url_entry, links->href);
		  i = parse_url (links->href, paula);
		  if (i < 0)
		    fubar (NONSYS, "parse_url died?  I'm not setup to cope with this.  I have _issues_.");
		  body = get_html_file (paula);
		  parse_html (body, paula);
		  free (body);
		  return TRUE;
		}
	    }
	  else
	    debug (debug_lisa, "no match with %fx%d in \"%s\" at %dx%d.%d+%d (%d)",
		   event->x, real_y, (links->href) ? links->href : "",
		   links->area->x, links->area->y,
		   links->area->width, links->area->height, lisa->virtual_y_offset);


	  links = links->next;
	}
      return TRUE;
    }
  debug (debug_lisa, "gtk_lisa_button_press: unsupported action (%d)", event->button);

  return FALSE;
}

/* Handle keypresses.
 *
 * I'm not entirely sure a) where keypresses are coming from and b) if
 * I'm seeing all the keypresses I should be.  Hmm. */

static gint
gtk_lisa_key_press (GtkWidget * widget,
		    GdkEventKey * event)
{

  GtkLisa *lisa;
  char key;
  int return_val;

  debug (debug_lisa, "gtk_lisa_key_press: got a key press");

  if (!widget)
    return FALSE;
  if (!GTK_IS_LISA (widget))
    return FALSE;
  if (!event)
    return FALSE;

  lisa = GTK_LISA (widget);
  key = event->keyval;
  return_val = TRUE;

  switch (event->keyval)
    {
    case GDK_Home:
      if (event->state & GDK_CONTROL_MASK)
	scroll_int (lisa, -lisa->vadj->value);
      break;
    case GDK_End:
      if (event->state & GDK_CONTROL_MASK)
	scroll_int (lisa, +lisa->vadj->upper);
      break;
    case GDK_Page_Up:
      scroll_int (lisa, -lisa->vadj->page_increment);
      break;
    case GDK_Page_Down:
      scroll_int (lisa, +lisa->vadj->page_increment);
      break;
    case GDK_Up:
      scroll_int (lisa, -KEY_SCROLL_PIXELS);
      break;
    case GDK_Down:
      scroll_int (lisa, +KEY_SCROLL_PIXELS);
      break;
    case GDK_Return:
      if (event->state & GDK_CONTROL_MASK)
	gtk_signal_emit_by_name (GTK_OBJECT (lisa), "activate");
      break;
    default:
      return_val = FALSE;
      break;
    }

  return return_val;
}

void
lisa_set_colours (GtkLisa *lisa, const colour_type *background_colour, 
		  const colour_type *link_colour, const colour_type *visited_link_colour, 
		  const colour_type *active_link_colour)
{

  GdkColormap *cmap;

  if (background_colour && !lisa->background_colour_set)
    {
      GdkColor *gdk_background_colour;

      gdk_background_colour = xmalloc (sizeof (*gdk_background_colour));
      gdk_background_colour->red = background_colour->red * 256;
      gdk_background_colour->green = background_colour->green * 256;
      gdk_background_colour->blue = background_colour->blue * 256;
      cmap = gtk_widget_get_colormap (&lisa->widget);
      if (!gdk_color_alloc (cmap, gdk_background_colour))
	fubar (NONSYS, "lisa_set_colours: couldn't allocate colour map");
      gdk_window_set_background (lisa->nicole, gdk_background_colour);
      free (gdk_background_colour);
    }

  if (link_colour && !lisa->link_colour)
    {
      lisa->link_colour = xmalloc (sizeof (*lisa->link_colour));
      lisa->link_colour->red = link_colour->red * 256;
      lisa->link_colour->green = link_colour->green * 256;
      lisa->link_colour->blue = link_colour->blue * 256;
      debug (debug_lisa, "Jeri: link = %x%x%x", link_colour->red, link_colour->green, link_colour->blue);
    }

  if (visited_link_colour && !lisa->visited_link_colour)
    {
      lisa->visited_link_colour = xmalloc (sizeof (*lisa->visited_link_colour));
      lisa->visited_link_colour->red = visited_link_colour->red * 256;
      lisa->visited_link_colour->green = visited_link_colour->green * 256;
      lisa->visited_link_colour->blue = visited_link_colour->blue * 256;
    }

  if (active_link_colour && !lisa->active_link_colour)
    {
      lisa->active_link_colour = xmalloc (sizeof (*lisa->active_link_colour));
      lisa->active_link_colour->red = active_link_colour->red * 256;
      lisa->active_link_colour->green = active_link_colour->green * 256;
      lisa->active_link_colour->blue = active_link_colour->blue * 256;
    }

}
