/* YAWN, Yet Another Web Navigator */
/* Copyright (C) 1998, James Troup <jjtroup@comp.brad.ac.uk>
 *                     Lee Willis <lwillis@comp.brad.ac.uk> */

/* This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#ifndef __LISA_H__
#define __LISA_H__

#include <gdk/gdk.h>
#include <gtk/gtkadjustment.h>
#include <gtk/gtkwidget.h>
#include "neve.h"

#define GTK_LISA(obj)          GTK_CHECK_CAST (obj, gtk_lisa_get_type (), GtkLisa)
#define GTK_LISA_CLASS(klass)  GTK_CHECK_CLASS_CAST (klass, gtk_lisa_get_type (), GtkLisaClass)
#define GTK_IS_LISA(obj)       GTK_CHECK_TYPE (obj, gtk_lisa_get_type ())


typedef struct _GtkLisa GtkLisa;
typedef struct _GtkLisaClass GtkLisaClass;

struct _GtkLisa
  {
    GtkWidget widget;
    GdkWindow *nicole;

    GtkAdjustment *hadj;
    GtkAdjustment *vadj;

    GdkGC *gc;

    /* For scrolling */

    gint last_ver_value;

    /* Button currently pressed or 0 if none */
    guint8 button;

    /* Offset between Y of virtual page and Y on screen */
    int offset;

    /* Current Y coordinate into the virtual page */
    int virtual_y_offset;

    /* The neve struct associated with this lisa */
    neve_in_chains *neve;

    /* The current line cache for the neve */
    louise *line_cache;

    /* Linked list of links */
    charisma *links;

    /* Are we in a link right now?

     * FIXME: this is possibly not the right location for this
     * variable, however having it as a global with file scope in
     * neve.c did not appeal. */

    boolean in_a_link;

    /* Colours */

    boolean background_colour_set;
    GdkColor *link_colour;
    GdkColor *visited_link_colour;
    GdkColor *active_link_colour;

  };

struct _GtkLisaClass
  {
    GtkWidgetClass parent_class;
  };


GtkWidget *gtk_lisa_new (GtkAdjustment * hadj, GtkAdjustment * vadj);
guint gtk_lisa_get_type (void);
void lisa_set_colours (GtkLisa *lisa, const colour_type *background_colour, 
		       const colour_type *link_colour, const colour_type *visited_link_colour, 
		       const colour_type *active_link_colour);

#endif /* __LISA_H__ */
