/* YAWN, Yet Another Web Navigator */
/* Copyright (C) 1998, James Troup <jjtroup@comp.brad.ac.uk>
 *                     Lee Willis <lwillis@comp.brad.ac.uk> */

/* This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

/** lisa_expose - Code for expose handling **/


#include <gdk/gdk.h>
#include "utils.h"
#include "render.h"
#include "neve.h"
#include "lisa.h"
#include "lisa_expose.h"
#include "lisa_scrolling.h"

/* Expose handler as called by other parts of Gtk+ */

gint
gtk_lisa_expose (GtkWidget * widget, GdkEventExpose * event)
{

  /* Standard sanity checks */

  if (!widget)
    return FALSE;
  if (!GTK_IS_LISA (widget))
    return FALSE;
  if (!event)
    return FALSE;
  if (event->count > 0)
    return FALSE;

  process_expose (GTK_LISA (widget), &event->area, event->area.y);

  return FALSE;
}

/* Expose handler as called by other parts of my code */

void
process_expose (GtkLisa * lisa, GdkRectangle * area, int real_y)
{

  louise *line_cache;
  int offset, end_of_window, diff;

  offset = lisa->offset;
  end_of_window = real_y + area->height;

  debug (debug_lisa_expose, "gtk_lisa_expose: x=%d y=%d w=%d h=%d, offset=%d real y=%d",
	 area->x, area->y, area->width, area->height, offset, real_y);

  line_cache = lisa->line_cache;
  if (line_cache)
    debug (debug_lisa_expose, "gtk_lisa_expose: line_cache is non-null");

  /* Find start in line cache */

  while (line_cache->y_coordinate < area->y && line_cache->next)
    line_cache = line_cache->next;

  /* Skip back one, to compensate for area->y's in the middle of a line */

  if (line_cache->previous)
    line_cache = line_cache->previous;

  debug (debug_lisa_expose, "gtk_lisa_expose: found start at %d (ln %d) (looking for %d)", 
	 line_cache->y_coordinate, line_cache->line_number, area->y);

  /* FIXME: what does this code do? (Yes, I did write it, I just won't remember) */

  if (line_cache->y_coordinate < area->y)
    {
      diff = area->y - line_cache->y_coordinate;
      area->y -= diff;
      real_y -= diff;
      area->height += diff;
      end_of_window = real_y + area->height;
    }

  debug (debug_lisa_expose, "gtk_lisa_expose: clearing %dx%d+%d.%d (%d+%d)", 
	 area->x, real_y + offset, area->width, area->height, real_y, offset);

  gdk_window_clear_area (lisa->nicole,
			 area->x, real_y + offset,
			 area->width, area->height);

  /* Render from line cache till y_coordinate > height... */

  debug (debug_lisa_expose, "gtk_lisa_expose[3]: %d vs %d", line_cache->y_coordinate, end_of_window);

  while (real_y < end_of_window && line_cache->next)
    {
      debug (debug_lisa_expose, "gtk_lisa_expose[2]: %d < %d (%d + %d) [offset is %d]", line_cache->y_coordinate, end_of_window, real_y, area->height, offset);
      render_line (line_cache, lisa, area, &real_y);
      line_cache = line_cache->next;
    }

  adjust_adj (lisa, lisa->vadj);

}
