/* YAWN, Yet Another Web Navigator */
/* Copyright (C) 1998, James Troup <jjtroup@comp.brad.ac.uk>
 *                     Lee Willis <lwillis@comp.brad.ac.uk> */

/* This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

/** lisa_scrolling - Code for adjustments and scrolling **/

#include <stdio.h>
#include <gtk/gtkmain.h>
#include <gtk/gtksignal.h>
#include <gdk/gdk.h>
#include "utils.h"
#include "neve.h"
#include "lisa.h"
#include "lisa_scrolling.h"
#include "lisa_expose.h"

void
gtk_lisa_set_adjustments (GtkLisa * lisa,
			  GtkAdjustment * hadj,
			  GtkAdjustment * vadj)
{

  if (!lisa)
    return;
  if (!GTK_IS_LISA (lisa))
    return;

  if (lisa->hadj && (lisa->hadj != hadj))
    {
      gtk_signal_disconnect_by_data (GTK_OBJECT (lisa->hadj), lisa);
      gtk_object_unref (GTK_OBJECT (lisa->hadj));
    }

  if (lisa->vadj && (lisa->vadj != vadj))
    {
      gtk_signal_disconnect_by_data (GTK_OBJECT (lisa->vadj), lisa);
      gtk_object_unref (GTK_OBJECT (lisa->vadj));
    }

  if (!hadj)
    hadj = GTK_ADJUSTMENT (gtk_adjustment_new (0.0, 0.0, 0.0, 0.0, 0.0, 0.0));

  if (!vadj)
    vadj = GTK_ADJUSTMENT (gtk_adjustment_new (0.0, 0.0, 0.0, 0.0, 0.0, 0.0));

  if (lisa->hadj != hadj)
    {
      lisa->hadj = hadj;
      gtk_object_ref (GTK_OBJECT (lisa->hadj));
      gtk_object_sink (GTK_OBJECT (lisa->hadj));

      gtk_signal_connect (GTK_OBJECT (lisa->hadj), "changed",
			  (GtkSignalFunc) gtk_lisa_adjustment,
			  lisa);
      gtk_signal_connect (GTK_OBJECT (lisa->hadj), "value_changed",
			  (GtkSignalFunc) gtk_lisa_adjustment,
			  lisa);
      gtk_signal_connect (GTK_OBJECT (lisa->hadj), "disconnect",
			  (GtkSignalFunc) gtk_lisa_disconnect,
			  lisa);
    }

  if (lisa->vadj != vadj)
    {
      lisa->vadj = vadj;
      gtk_object_ref (GTK_OBJECT (lisa->vadj));
      gtk_object_sink (GTK_OBJECT (lisa->vadj));

      gtk_signal_connect (GTK_OBJECT (lisa->vadj), "changed",
			  (GtkSignalFunc) gtk_lisa_adjustment,
			  lisa);
      gtk_signal_connect (GTK_OBJECT (lisa->vadj), "value_changed",
			  (GtkSignalFunc) gtk_lisa_adjustment,
			  lisa);
      gtk_signal_connect (GTK_OBJECT (lisa->vadj), "disconnect",
			  (GtkSignalFunc) gtk_lisa_disconnect,
			  lisa);
    }

}

void
gtk_lisa_adjustment (GtkAdjustment * adjustment,
		     GtkLisa * lisa)
{

  if (!adjustment)
    return;
  if (!GTK_IS_ADJUSTMENT (adjustment))
    return;
  if (!lisa)
    return;
  if (!GTK_IS_LISA (lisa))
    return;

  if (adjustment == lisa->hadj)
    {
      debug (debug_warn, "gtk_lisa_adjustment: horizontal scrolling not implemented");
    }
  else
    {
      gint diff = ((gint) adjustment->value) - lisa->last_ver_value;

      if (diff != 0)
	{

	  if (diff > 0)
	    scroll_down (lisa, diff);
	  else	/* if (diff < 0) */
	    scroll_up (lisa, diff);

	  lisa->last_ver_value = adjustment->value;
	}
    }
}

void
gtk_lisa_disconnect (GtkAdjustment * adjustment,
		     GtkLisa * lisa)
{
  if (!adjustment)
    return;
  if (!GTK_IS_ADJUSTMENT (adjustment))
    return;
  if (!lisa)
    return;
  if (!GTK_IS_LISA (lisa))
    return;

  if (adjustment == lisa->hadj)
    lisa->hadj = NULL;
  if (adjustment == lisa->vadj)
    lisa->vadj = NULL;
}

void
adjust_adj (GtkLisa * lisa, GtkAdjustment * adj)
{
  gint height;

  gdk_window_get_size (lisa->nicole, NULL, &height);

  adj->step_increment = MIN (adj->upper, (float) SCROLL_PIXELS);
  adj->page_increment = MIN (adj->upper, height - (float) KEY_SCROLL_PIXELS);
  adj->page_size = MIN (adj->upper, height);
  adj->value = MIN (adj->value, adj->upper - adj->page_size);
  adj->value = MAX (adj->value, 0.0);

  gtk_signal_emit_by_name (GTK_OBJECT (adj), "changed");
}

void
scroll_int (GtkLisa * lisa, int diff)
{
  gfloat upper;

  lisa->vadj->value += diff;

  upper = lisa->vadj->upper - lisa->vadj->page_size;
  lisa->vadj->value = MIN (lisa->vadj->value, upper);
  lisa->vadj->value = MAX (lisa->vadj->value, 0.0);

  debug (debug_lisa_scroll, "scroll_int: scrolling by %d", diff);

  gtk_signal_emit_by_name (GTK_OBJECT (lisa->vadj), "value_changed");
}

/* FIXME: this doesn't work 100% */

void
scroll_down (GtkLisa * lisa, int diff)
{

  static boolean are_we_a_double = FALSE; 
  gint width, height;
  GdkRectangle render_area;
  int real_y;
  gdk_window_get_size (lisa->nicole, &width, &height);

  if (are_we_a_double)
    {
      debug (debug_lisa_scroll, "scroll_down: double; outie.");
      are_we_a_double = FALSE;
      return;
    }
  else 
    are_we_a_double = TRUE;

  debug (debug_lisa_scroll, "scroll_down[1]: starting...");
  debug (debug_lisa_scroll, "scroll_down: current y is %d", lisa->virtual_y_offset);

  render_area.x = 0;
  render_area.width = width;
  render_area.height = diff;
  lisa->virtual_y_offset += diff;

  debug (debug_lisa_scroll, "scroll_down:(2) current y is %d", lisa->virtual_y_offset);

  if (diff >= height)
    {  /* We need to rerender the entire real world window */
      debug (debug_lisa_scroll, "scroll_down: %d >= %d; can't blit, virtual_y_offset => %d", 
	     diff, height, lisa->virtual_y_offset);
      render_area.y = lisa->virtual_y_offset;
      real_y = 0;
    }
  else /* We can blit some of the screen and render only a portion */
    {
      debug (debug_lisa_scroll, "scroll_down:(2.5) height is %d, diff is %d, height-diff => is %d",
	     height, diff, height - diff);
      debug (debug_lisa_scroll, "scroll_down:(3) current y is %d", lisa->virtual_y_offset);

      gdk_draw_pixmap (lisa->nicole,
		       lisa->gc,
		       lisa->nicole,
		       0, 0 + diff,
		       0, 0, width, height - diff);
      real_y = (height - diff);
      render_area.y = lisa->virtual_y_offset + (height - diff);

      debug (debug_lisa_scroll, "scroll_down: %d < %d; current y => %d; render y => %d",
	     diff, height, lisa->virtual_y_offset, render_area.y);
      debug (debug_lisa_scroll, "scroll_down: blitted: %dx%d -> %dx%d  %d.%d", 
	     0, 0 + diff, 0, 0, width, height - diff);
    }

  process_expose (lisa, &render_area, real_y);

}

/* FIXME: a) this is as not 100% as scroll_down, b) duplicates
   scroll_up, most horribly */

void
scroll_up (GtkLisa * lisa, int diff)
{

  static boolean are_we_a_double = FALSE; 
  gint width, height;
  GdkRectangle render_area;
  int real_y;
  gdk_window_get_size (lisa->nicole, &width, &height);

  if (are_we_a_double)
    {
      debug (debug_lisa_scroll, "scroll_down: double; outie.");
      are_we_a_double = FALSE;
      return;
    }
  else 
    are_we_a_double = TRUE;
  /* No negative differences please, otherwise the logic gets
     contorted (or is that _more_ contorted?) */

  diff = abs (diff);

  debug (debug_lisa_scroll, "scroll_up[1]: starting...");
  debug (debug_lisa_scroll, "scroll_up: current y is %d", lisa->virtual_y_offset);

  render_area.x = 0;
  render_area.width = width;
  render_area.height = diff;
  lisa->virtual_y_offset -= diff;

  debug (debug_lisa_scroll, "scroll_up:(2) current y is %d", lisa->virtual_y_offset);

  if (diff >= height)
    {
      debug (debug_lisa_scroll, "scroll_up: %d >= %d; can't blit, virtual_y_offset => %d", 
	     diff, height, lisa->virtual_y_offset);
      render_area.y = lisa->virtual_y_offset;
      real_y = 0;
    }
  else
    {
      debug (debug_lisa_scroll, "scroll_up:(2.5) height is %d, diff is %d, height-diff => is %d",
	     height, diff, height - diff);
      debug (debug_lisa_scroll, "scroll_up:(3) current y is %d", lisa->virtual_y_offset);

      gdk_draw_pixmap (lisa->nicole,
		       lisa->gc,
		       lisa->nicole,
		       0, 0,
		       0, 0 + diff, width, height - diff);
      real_y = 0;
      render_area.y = lisa->virtual_y_offset;	

      debug (debug_lisa_scroll, "scroll_up: %d < %d; current y => %d; render y => %d",
	     diff, height, lisa->virtual_y_offset, render_area.y);
      debug (debug_lisa_scroll, "scroll_up: blitted: %dx%d -> %dx%d  %d.%d",
	     0, 0, 0, 0 + diff, width, height - diff);
    }

  process_expose (lisa, &render_area, real_y);

}
