/* YAWN, Yet Another Web Navigator */
/* Copyright (C) 1998, James Troup <jjtroup@comp.brad.ac.uk>
 *                     Lee Willis <lwillis@comp.brad.ac.uk> */

/* This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include <stdio.h>
#include <string.h>
#include <gtk/gtkwidget.h>
#include "neve.h"
#include "neve_misc.h"
#include "utils.h"
#include "render.h"
#include "lisa.h"
#include "smg.h"
#include "lisa_expose.h"
#include "alyson.h"

/* Declarations for private functions */

void neve_to_louise (neve_in_chains * neve, GtkLisa * lisa);
louise_parse_neve_return_type louise_parse_neve (GtkLisa * lisa, neve_in_chains * neve,
						 charisma ** links, const int nicole_width,
						 int *x_coordinate, const int y_coordinate,
						 int *max_font_height, int *offset_into_neve, 
						 boolean first_neve);
louise_parse_neve_return_type louise_parse_text (GtkLisa * lisa, neve_in_chains * neve,
						 charisma ** links, const int nicole_width,
						 int *x_coordinate, const int y_coordinate,
						 int *max_font_height, int *offset);
void update_charisma (GtkLisa * lisa, charisma ** current_link, const char *href,
		      const int x_coordinate, const int y_coordinate, const int width,
		      const int height, boolean line_ending);

/*****************************************************************************
 * function  : lisa_take_neve
 * arguments : A neve and a lisa
 * results   : 
 * purpose   : takes neve from the parser and calls the neve2louise convertor.
 * caveats   : 
 ***************************************************************************** */

void
lisa_take_neve (neve_in_chains * neve, BrowserWindow *paula)
{
  
  GtkLisa *lisa;
  neve_in_chains *temp;
  boolean redraw = FALSE;
  gint width, height;
  GdkRectangle render_area;

  lisa = GTK_LISA(paula->lisa);
  temp = neve;

  if (lisa->neve)
    {
      debug (debug_neve, "lisa_take_neve: free-ing previous neve struct");
      free_neve (lisa->neve);
      free_charisma (lisa);
      redraw = TRUE;
    }

  lisa->neve = neve;
  neve_to_louise (lisa->neve, lisa);

  if (!lisa->link_colour)
    {
      colour_type link_colour;
      
      link_colour.red = 255;
      link_colour.green = 0;
      link_colour.blue = 255;
      lisa_set_colours (lisa, NULL, &link_colour, NULL, NULL);
    }

  if (redraw)
    {
      gdk_window_get_size (lisa->nicole, &width, &height);
      gdk_window_clear_area (lisa->nicole,
			     0, 0,
			     width, height);
      render_area.x = 0;
      render_area.y = 0;
      render_area.width = width;
      render_area.height = height;
      process_expose (lisa, &render_area, 0);
    }

}

/*****************************************************************************
 * function  : neve_to_louise
 * arguments : A neve and a lisa
 * results   : 
 * purpose   : takes a neve and creates a suitable louise line cache from it.
 * caveats   : buggy about Y coordinates, multi-line list elements
 ***************************************************************************** */

void
neve_to_louise (neve_in_chains * neve, GtkLisa * lisa)
{

  int font_height, line_number, x_coordinate, y_coordinate, offset, line_space;
  louise_parse_neve_return_type i;
  charisma *links = NULL;
  gint nicole_width;
  louise *current_line = NULL;
  louise *previous_line = NULL;
  boolean first_neve = TRUE; /* First neve in line */
  neve_in_chains *line_break = NULL;

  gdk_window_get_size (lisa->nicole, &nicole_width, NULL);
  nicole_width -= 4;
  font_height = 0;
  offset = 0;
  y_coordinate = 0;
  line_number = 1;
  line_space = 2;

  current_line = xmalloc (sizeof (*current_line));
  current_line->start_of_line = neve;
  current_line->offset_into_first_neve = 0;
  current_line->line_number = line_number;
  current_line->y_coordinate = y_coordinate;
  current_line->max_font_height = 0;
  current_line->next = NULL;
  current_line->previous = NULL;

  lisa->line_cache = current_line;
  lisa->in_a_link = FALSE;

  line_number++;

  while (neve)
    {
      debug (debug_neve, "neve_to_louise: doing line %d", line_number);
      debug (debug_neve, "neve_to_louise: and y_coordinate = %d", y_coordinate);
      x_coordinate = 0;
      first_neve = TRUE;
      do
	{
	  font_height = 0;
	  debug (debug_neve, "neve_to_louise: searching... (%dx%d, %d, %d)",
		 x_coordinate, y_coordinate, font_height, offset);
	  i = louise_parse_neve (lisa, neve, &links, nicole_width, &x_coordinate,
				 y_coordinate, &font_height, &offset, first_neve);
	  if (font_height > current_line->max_font_height)
	    {
	      debug (debug_neve, "neve_to_louise: max_font_height => %d", font_height);
	      current_line->max_font_height = font_height;
	    }
	  first_neve = FALSE;
	  if (i == exhausted_node || i == skip_to_next_line)
	    {
	      debug (debug_neve, "neve_to_louise: moving onto next neve node");
	      neve = neve->next;
	      offset = 0;
	    }

	  if (i == skip_to_next_paragraph)
	    {
	      debug (debug_neve, "neve_to_louise: skipping a line");
	      line_break = neve;
	      neve = neve->next;
	    }

	  if (i == hit_end_of_line)
	    debug (debug_neve, "neve_to_louise: we hit the end of the line d00d!  Game Over, man.");
	  if (i == skip_to_next_paragraph)
	    {
	      debug (debug_neve, "neve_to_louise: skipidi do da skipidi day.");
	    }
	  if (i == skip_to_next_line)
	    {
	      debug (debug_neve, "neve_to_louise: skip to next line.");
	    }
	}
      while (i == exhausted_node && neve);

      /* If max_font_height is still 0, we must have had a line or
       * paragraph break, adjust. */
      if (current_line->max_font_height == 0)
	current_line->max_font_height = 6;
      
      font_height = current_line->max_font_height;

      if (line_break)
	{
	  debug (debug_neve, "compensating for a line break");

	  previous_line = current_line;
	  current_line = xmalloc (sizeof (*current_line));
	  current_line->start_of_line = line_break;
	  current_line->offset_into_first_neve = 0;
	  current_line->line_number = line_number;
	  current_line->y_coordinate = y_coordinate;
	  current_line->max_font_height = 6;
	  current_line->next = NULL;
	  current_line->previous = previous_line;
	  previous_line->next = current_line;

	  line_number++;
	  debug (debug_neve, "neve_to_louise: y_coordinate (%d) += %d + %d (%d)", y_coordinate, current_line->max_font_height, line_space, line_space + current_line->max_font_height);
	  y_coordinate += current_line->max_font_height + line_space;

	}

      line_break = NULL;

      previous_line = current_line;
      current_line = xmalloc (sizeof (*current_line));
      current_line->start_of_line = neve;
      current_line->offset_into_first_neve = offset;
      current_line->line_number = line_number;
      current_line->y_coordinate = y_coordinate;
      current_line->max_font_height = 0;
      current_line->next = NULL;
      current_line->previous = previous_line;
      previous_line->next = current_line;

      line_number++;
      debug (debug_neve, "neve_to_louise: y_coordinate (%d) += %d + %d (%d)", y_coordinate, font_height, line_space, line_space + font_height);
      y_coordinate += font_height + line_space;

      debug (debug_neve, "neve_to_louise: y_coordinate = %d", y_coordinate);

    }

  debug (debug_neve, "neve_to_louise: setting lisa->vadj->upper to %d", current_line->y_coordinate);
  lisa->vadj->upper = current_line->y_coordinate;

}

/* This procedure is called by neve2louise() for each neve node.  It
 * simply switches on the type of the neve node and calls functions
 * accordingly. */

louise_parse_neve_return_type
louise_parse_neve (GtkLisa * lisa, neve_in_chains * neve, charisma ** links, const int nicole_width,
		   int *x_coordinate, const int y_coordinate, int *max_font_height, 
		   int *offset_into_neve, boolean first_neve)
{

  int return_value;

  if (first_neve && neve->logical_attributes.list_indent_level > 0)
    (*x_coordinate) += 25 + (neve->logical_attributes.list_indent_level * 10);

  switch (neve->type)
    {
    case text:
      debug (debug_neve, "louise_parse_neve: found text.");
      return_value = louise_parse_text (lisa, neve, &(*links), nicole_width, &(*x_coordinate),
			     y_coordinate, &(*max_font_height), &(*offset_into_neve));
      break;
    case image:
      debug (debug_neve, "louise_parse_neve: found image");
      return_value = louise_parse_text (lisa, neve, &(*links), nicole_width, &(*x_coordinate),
			     y_coordinate, &(*max_font_height), &(*offset_into_neve));
      break;
    case horizontal_rule:
      debug (debug_neve, "louise_parse_neve: found horizontal rule.");
      return_value = skip_to_next_line;
      break;
    case line_break:
      debug (debug_neve, "louise_parse_neve: found line break.");
      return_value = skip_to_next_line;
      break;
    case paragraph_break:
      debug (debug_neve, "louise_parse_neve: found paragraph break.");
      return_value = skip_to_next_paragraph;
      break;
    case end_link:
      debug (debug_neve, "louise_parse_neve: found end_link.");
      lisa->in_a_link = FALSE;
      return_value = exhausted_node;
      break;
    case start_list_element:
      debug (debug_neve, "louise_parse_neve: found start of list.");
      return_value = exhausted_node;
      break;
    default:
      fubar (NONSYS, "louise_parse_neve: found %d, which I clearly don't recognise. outie :-(.",
	     neve->type);
    }

  return return_value;

}

/* Generic text handling.  There is a *lot* of code duplication with
 * the render, since this is doing everything the render does, except
 * draw it.  It would _really_ be a good plan to merge these two, and
 * have the actual rendering conditional on an argument. 
 *
 * Also the number of arguments is plainly silly, and could do with
 * some rationalization, somehow. */

louise_parse_neve_return_type
louise_parse_text (GtkLisa * lisa, neve_in_chains * neve, charisma ** links, const int nicole_width,
		   int *x_coordinate, const int y_coordinate, int *max_font_height, int *offset_into_neve)
{

  int start_of_link = *x_coordinate;
  int start_of_region = *offset_into_neve;
  int end_of_region = start_of_region;
  louise_parse_neve_return_type return_value;
  char fontname[256];
  GdkFont *font;
  int font_size, font_height;

  int i;
  char foo[10];


  for (i = 0; i < 8; i++)
    {
      if (neve->buf[*offset_into_neve+i] == '\0')
	break;
      else
	foo[i] = neve->buf[*offset_into_neve+i];
    }
  foo[i] = '\0';

  debug (debug_neve, "louise_parse_text: *:%s:* [%d: offset_into_neve %d], start of region = %d, end = %d",
	 foo, strlen (neve->buf), *offset_into_neve, start_of_region, end_of_region);
  debug (debug_neve, "louise_parse_text: x_coordinate = %d, start_of_region = %d", *x_coordinate, start_of_region);

  debug (debug_render, "render_text: font level is %d", neve->physical_attributes.font_level);
  font_size = font_levels[neve->physical_attributes.font_level - 1];
  debug (debug_render, "render_text: font size --> %d", font_size);



  sprintf (fontname, "-adobe-%s-%s-%s-normal--%d-100-*-*-*-*-*-*",
	   neve->physical_attributes.monospace ? "courier" : "helvetica", 
	   neve->physical_attributes.strong ? "bold" : "medium",
  /* WARNING: abode-helvetica uses "o" not "i", mucking with
   * this can cause the X server to keel over.  Great, eh?
   */
	   neve->physical_attributes.emphasis ? "o" : "r",
	   font_size);
  debug (debug_neve, "louise_parse_text: fontname is \"%s\"", fontname);

  font = gdk_font_load (fontname);

  if (font == NULL)
    {
      fubar (NONSYS, "louise_parse_text: couldn't load font. outie");
    }

  debug (debug_neve, "linda: %d => %d + %d", font_size, font->ascent, font->descent);

  font_height = font->ascent + font->descent;
  if (font_height > *max_font_height)
    *max_font_height = font_size;

  while (*x_coordinate < nicole_width && neve->buf[end_of_region] != '\0')
    {
      (*x_coordinate) += gdk_char_width (font, neve->buf[end_of_region]);
      end_of_region++;
    }

  debug (debug_neve, "louise_parse_text: [%d >= %d:%d:%c]",
      *x_coordinate, nicole_width, end_of_region, neve->buf[end_of_region]);

  if (neve->buf[end_of_region] == '\0')		/* neve node is finished */
    {
      debug (debug_neve, "louise_parse_text: I think the neve buffer is finished.");
      return_value = exhausted_node;
    }
  else
    /* We've hit the end of line */
    {
      debug (debug_neve, "louise_parse_text: I think the neve buffer is NOT finished.");
      backup_to_last_whitespace (neve, start_of_region, &end_of_region, &(*x_coordinate), font);
      return_value = hit_end_of_line;
    }

  debug (debug_neve, "louise_parse_text: '%c' <-----", neve->buf[end_of_region]);
  /* FIXME: I think this could overflow */
  if (neve->buf[end_of_region] == ' ' && return_value == hit_end_of_line)
    {
      end_of_region++;
      debug (debug_neve, "louise_parse_text: '%c' <~~~~~", neve->buf[end_of_region]);
    }

  debug (debug_neve, "louise_parse_text: start of link = %d, start_of_region = %d, offset_into_neve is %d, end_of_region is %d",
	 start_of_link, start_of_region, *offset_into_neve, end_of_region);
  strncpy (foo, neve->buf + start_of_region, end_of_region - start_of_region);
  foo[end_of_region - start_of_region] = '\0';
  debug (debug_neve, "louise_parse_text: !:%s:!", foo);

  /* Handle links */

  if (neve->logical_attributes.link)
    {
      debug (debug_neve, "louise_parse_text: found link at %dx%d %d+%d [\"%s\"] => \"%s\"",
	     start_of_link, y_coordinate,
	     *x_coordinate - start_of_link, *max_font_height,
	     foo, (neve->href) ? neve->href : "");
      if (return_value == hit_end_of_line)
	debug (debug_neve, "this is quite clearly the end of the line");
      /* FIXME: I suck */
      update_charisma (lisa, &(*links), neve->href, start_of_link, y_coordinate + 15,
		       *x_coordinate - start_of_link, *max_font_height,
		       (return_value == hit_end_of_line) ? TRUE : FALSE);

    }

  *offset_into_neve = end_of_region;
  return return_value;

}

/* Add/update a link to the charisma list of links. */

void
update_charisma (GtkLisa * lisa, charisma ** current_link, const char *href,
		 const int x_coordinate, const int y_coordinate, const int width,
		 const int height, boolean line_ending)
{

  charisma *new_link;


  if (lisa->in_a_link)
    {
      (*current_link)->area->width += width;
      (*current_link)->area->height += height;
    }
  else
    /* !in a link */
    {
      new_link = xmalloc (sizeof (*new_link));
      if (*current_link)
	(*current_link)->next = new_link;
      *current_link = new_link;
      (*current_link)->area = xmalloc (sizeof (*new_link->area));
      (*current_link)->area->x = x_coordinate;
      (*current_link)->area->y = y_coordinate;
      (*current_link)->area->width = width;
      (*current_link)->area->height = height;
      (*current_link)->href = xmalloc (strlen (href) + 1);
      strcpy ((*current_link)->href, href);
      (*current_link)->next = NULL;
    }

  /* If the global links start pointer doesn't exist; initalize it. */
  if (!lisa->links)
    lisa->links = *current_link;

  /* Handle multi-line links.

   * Multi-line links have to be split into realisitic rectangles.  At
   * the moment this is done by creating a link for each line of a
   * link.  It could, possibly, be optimized to group lines of the
   * same width of one multi-line link into one big rectangle as
   * opposed to two or more seperate rectangles. */

  if (line_ending)
    {
      new_link = xmalloc (sizeof (*new_link));
      new_link->area = xmalloc (sizeof (*new_link->area));
      new_link->area->x = 0;
      /* FIXME: this is FUBAR. we need to take into account the max font height of the line */
      new_link->area->y = y_coordinate + height + 12;
      new_link->area->width = 0;
      new_link->area->height = 0;
      new_link->href = xmalloc (strlen ((*current_link)->href) + 1);
      strcpy (new_link->href, (*current_link)->href);
      new_link->next = NULL;
      (*current_link)->next = new_link;
      (*current_link) = new_link;
    }

  lisa->in_a_link = TRUE;

}
