/* YAWN, Yet Another Web Navigator */
/* Copyright (C) 1998, James Troup <jjtroup@comp.brad.ac.uk>
 *                     Lee Willis <lwillis@comp.brad.ac.uk> */

/* This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include <stdlib.h>
#include <string.h>
#include "latin.h"
#include "utils.h"


/* latin.c
 * -------
 *
 * parse_latin_entities accepts a pointer to a null terminated string
 * of characters and removes latin character entities of the form
 * &name; or &#xxx as defined in the HTML 3.2 standard (Available at
 * http://www.w3.org/TR/REC-html32.html#latin1) and replacing them with
 * their proper values before passing the string back to the calling
 * function.
 *
 * NB. Currently only supports &copy; &lt; &gt; &amp; and their &#xxx
 *     equivalents.
 */

void replace (char new_char, char *buffer, long *index, int length);
void replace (char new_char, char *buffer, long *index, int length)
{
  buffer[*index] = new_char; /* Replace the & with the new character */
  (*index)++; /* Skip the newly created character */
  debug ( debug_parser, "latin.c: Now pointing at->%s", buffer+(*index));
  if (buffer[*index+length]==';')
    strcpy(buffer+(*index), buffer+(*index)+length+1); /* Shift the rest of the text over the name; */
  else
    strcpy(buffer+(*index), buffer+(*index)+length); /* Shift the rest of the text over the name */
  debug ( debug_parser, "latin.c: Now pointing at->%s", buffer+(*index));
}



void parse_latin_entities(char *buffer)
{
  long index = 0;
  int entity_found;
  char *endofentity;

  while (buffer[index] != '\0')
    
    {

      if (buffer[index] == '&')

	{
	
	  debug ( debug_parser, "latin.c: Found a & in the string %s", buffer+index);
	  
	  if ((strncasecmp(buffer+index+1, "nbsp", 4) == 0) ||
	      (strncasecmp(buffer+index+1, "#160", 4) == 0))
	    replace ( ' ', buffer, &index, 4);
	  
	  else if (strncasecmp(buffer+index+1, "amp", 3) == 0 ||
		   strncmp(buffer+index+1, "#38", 3) == 0 )
	    replace ( '&', buffer, &index,3);
	  
	  else if (strncasecmp(buffer+index+1, "gt", 2) == 0 ||
		   strncmp(buffer+index+1, "#62", 3) == 0 )
	    replace ( '>', buffer, &index, buffer[index+1]=='#' ? 3 : 2);
	  
	  else if (strncasecmp(buffer+index+1, "lt", 2) == 0 ||
		   strncmp(buffer+index+1, "#60", 3) == 0 )
	    replace ( '<', buffer, &index, buffer[index+1]=='#' ? 3 : 2);
	  
	  else if (strncasecmp(buffer+index+1, "copy", 4) == 0 ||
		   strncmp(buffer+index+1, "#169", 4) == 0 )
	    replace ( 169, buffer, &index, 4 );
	  
	  else if (strncasecmp(buffer+index+1, "#32", 3) == 0)
	    replace ( 32, buffer, &index, 3);

	  else if (strncasecmp(buffer+index+1, "quot", 4) == 0)
	    replace ( '"', buffer, &index, 4);

	  else if (buffer[index+1] == '#')
	    {
	      debug (debug_parser, "latin: general # parsing ...");
	      entity_found = (int) strtol (buffer+index+2, &endofentity, 10);
	      debug (debug_parser, "latin: Enitity %d found.", entity_found);
	      if (entity_found != 0)
		{
		  buffer[index] = entity_found; /* Replace with the correct char */
		  if (*endofentity == ';')
		    strcpy(buffer+index+1, endofentity+1);
		  else
		    strcpy (buffer+index+1, endofentity);
		}
	      else
		index++;
	    }
	      
	  else
	    index++;
	  /* Unrecognized entities just rendered as they appear in the source */
	  
	}
      
      else
	index++; /* Skip onto the next bit of the text. */
	
    }

}
