/* YAWN, Yet Another Web Navigator */
/* Copyright (C) 1998, James Troup <jjtroup@comp.brad.ac.uk>
 *                     Lee Willis <lwillis@comp.brad.ac.uk> */

/* This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

/* WARNING: This code contains comments and code that may offend, if you are a worry
 * wart or one of those sad little people who complain to Ann Robinson about the use
 * of the word backside on TV then please feel free not to read on .... */


#include <stdio.h>
#include <stdlib.h>
#include "utils.h"
#include "neve.h"
#include "scanner.h"
#include "tagdefs.h"
#include "parser.h"
#include "parse_attribs.h"
#include "attrib_utils.h"
#include "neve_operations.h"
#include "staxx.h"
#include "text.h"
#include "string.h"

#define foobar (objfound=scan(html_doc, file_index, text_buffer)) 

logical_attribute_types current_lattributes; /* State information, see 
					      *	neve.h for more details. */
physical_attribute_types current_pattributes; /* State information */
char *current_href; /* State information for links */



void add_lbreaks (neve_in_chains *node)
{
  debug (debug_parser, "add_lbreaks: pointing at ...");
  out_neve_type (node);
  
  while (node && (node->type == end_link || node->type == start_list_element))
    {
      node = node->previous;
      debug (debug_parser, "add_lbreaks: pointing at ...");
      if (node)
	out_neve_type (node);
      else
	debug (debug_parser, "NULL!");
    }
  
  if (node)
    if (node->type != paragraph_break && node->type != line_break)
      {
	if (node->type == text)
	  {
	    if (node->buf[0] == ' ' && node->buf[1] == '\0')
	      {
		free (node->buf);
		debug (debug_parser, "add_lbreaks: Swallowing 'empty' text node.");
		node->type = line_break;
	      }
	    else
	      {
		debug (debug_parser, "add_lbreaks: Adding LB" );
		new_neve_node (line_break, &current_pattributes, &current_lattributes, current_href);
	      }
	  }
	else
	  {
	    debug (debug_parser, "add_lbreaks: Adding LB" );
	    new_neve_node(line_break, &current_pattributes, &current_lattributes, current_href);
	  }
      }
}




void add_pbreaks (neve_in_chains *node)
{
  debug (debug_parser, "add_pbreaks: pointing at ...");
  out_neve_type (node);

  while (node && (node->type == end_link || node->type == start_list_element))
    {
      node = node->previous;
      debug (debug_parser, "add_pbreaks: pointing at ...");
      if (node)
	out_neve_type (node);
      else
	debug (debug_parser, "NULL!");
    }

  if (node)
    if (node->type == line_break)
      {
	debug (debug_parser, "add_pbreaks: Converting LB to PB." );
	node->type = paragraph_break;
      }
    else if (node->type != paragraph_break)
      {
	if (node->type == text)
	  {
	    if (node->buf[0] == ' ' && node->buf[1] == '\0')
	      {
		free (node->buf);
		debug (debug_parser, "add_pbreaks: Swallowing 'empty' text node.");
		node->type = paragraph_break;
	      }
	    else
	      {
		debug (debug_parser, "add_pbreaks: Adding PB" );
		new_neve_node (paragraph_break, &current_pattributes, &current_lattributes, current_href);
	      }
	  }
	else
	  {
	    debug (debug_parser, "add_lbreaks: Adding PB" );
	    new_neve_node(paragraph_break, &current_pattributes, &current_lattributes, current_href);
	  }
      }
}




int expect_document (BrowserWindow *window, char *html_doc, long *file_index, char *text_buffer)
{
  /* This is the main parsing function  that calls the scanner and the
   * routines to parse the attributes. As it goes along it creates
   * a list of neve nodes with the correct (Hopefully) values to be able
   * to pass the renderer a decent representation of the page.
   */
  
  int hlevel = 0; /* Variable used to construct the header level */
  int objfound = UNRECOGNIZED; /* Used to hold the code of the tag we've found */
  int previous_tag = UNRECOGNIZED; /* Used to hold the previous tag matched */
  int current_basefont_size = current_pattributes.font_level; /* Stores the current basefont size */
  int dl_top_indent_level = 0; /* Offset to add to get the indent level for dd's */
  boolean title_flag = FALSE;

  while ((objfound=scan(html_doc, file_index, text_buffer, current_lattributes.preformatted, previous_tag)) != END_OF_DOC) {

    hlevel = 0;

    debug (debug_parser, "parser: recieved a %d", objfound);

    if (objfound >= CLOSETAG && previous_tag == TEXT && neve_list && neve_list->type == text)
      {
	debug (debug_parser, "parser: neve_list points at %p", neve_list);
	if (neve_list)
	  {
	    int buffer_p = strlen (neve_list->buf) - 1; /* Point at last character */
	    while (neve_list->buf[buffer_p] == '\n' && buffer_p > 0) /* Remove new_lines preceeding close tags */
	      neve_list->buf[buffer_p--] = '\0'; 
	  }
      }
    
    if (previous_tag == TEXT && neve_list && neve_list->type == text)
      {
	debug ( debug_parser, "Folding text $$%s$$", neve_list->buf);
	fold (neve_list->buf);
	if (neve_list->buf[0]==' ')
	  {
	    debug (debug_parser, "Possibly folding spaces into the void ...");
	    { /* Scan back to the previous renderable node. */
	      neve_in_chains *temp = neve_list->previous;
	      while (temp && (temp->type == end_link || temp->type == start_list_element))
		temp = temp->previous;
	      debug (debug_parser, "Previous renderable node is type");
	      out_neve_type(temp);
	      if (!temp ||
		  temp->type == paragraph_break || 
		  temp->type == line_break ||
		  temp->type == horizontal_rule)
		{
		  if (neve_list->buf[1] =='\0')
		    {
		      debug (debug_parser, "Folding into the void ...");
		      free (neve_list->buf);
		      if (neve_list->previous)
			{
			  neve_list = neve_list->previous;
			  free (neve_list->next);
			  neve_list->next = NULL;
			}
		      else
			{
			  free (neve_list);
			  neve_list = NULL;
			  neve_head = NULL;
			}
		    }
		  else
		    {
		      debug (debug_parser, "Folding out preceeding space.");
		      strcpy (neve_list->buf, neve_list->buf+1);
		    }
		}
	    }
	    
	  }
      }
    
    
    if (title_flag && objfound < COMMENT)
      title_flag = FALSE;

    
    switch (objfound)
      {
      case TEXT:
	/* Handle plain text */
	debug (debug_parser, "parser: found text (Length %d)- %s", strlen(text_buffer), text_buffer);
        if (title_flag)
	  {
	    debug (debug_parser, "Setting title ...");
	    paula_set_title (window, text_buffer);
	    title_flag = FALSE;
	  }
	else
	  {
	    debug ( debug_parser, "Creating a TEXT node ...");
	    new_neve_node(text, &current_pattributes, &current_lattributes, current_href);
	    neve_list->buf = xmalloc(strlen(text_buffer)+1);
	    strcpy ( neve_list->buf, text_buffer);
	    text_buffer [0] = '\0';
	  }
	
	break;

	/* Note for BOLD, ITALIC etc.  These should worry about proper
	 * nesting of tags however bad nesting is such a common
	 * feature that we have simply gone for the idea that <B> sets
	 * the state to bold, until a </B> is found regardless of any
	 * intervening tags. ie
	 * 
	 * <B>Foo<I>bar</B>baz</I>bat
	 *
	 * will be rendered with the foo in bold, the bar in bold
	 * italics, the baz in italics and the bat in plain text as
	 * opposed to the standard which would enforce proper nesting
	 * ie <B><I></I></B> */

	/* We could probably have a generic open and close for bold,
	 * italic, underline, strike, underline, subscript,
	 * superscript etc instead of this hideous code duplication
	 * but it doesn't really matter for now.  */

      case TR:

	debug (debug_parser, "parser: Found <TR>");
	add_lbreaks(neve_list);
	skip_attribs(html_doc, file_index);
	break;	

      case TD|CLOSETAG:

	debug (debug_parser, "parser: Found <TD>");
	add_pbreaks(neve_list);
	skip_attribs(html_doc, file_index);
	break;

      case BOLD:

	debug (debug_parser, "parser: Found <B>");
	current_pattributes.strong = TRUE; /*Set the current state to reflect bold */
	skip_attribs(html_doc, file_index);
	break;
	
      case BOLD|CLOSETAG:

	debug (debug_parser, "parser: Found </B>");
	current_pattributes.strong = FALSE; /* Reset the bold state */
	skip_attribs(html_doc, file_index);
	break;


      case CENTER:
	
	debug (debug_parser, "parser: Found <CENTER>");
	push_align (current_pattributes.alignment, DIV); /* Push current alignment onto stack */
	current_pattributes.alignment = centre;
	skip_attribs(html_doc, file_index);
	break;
	

      case CENTER|CLOSETAG:

	debug(debug_parser, "parser: Found </CENTER>");
	pop_align(&current_pattributes, DIV); /* Pop last alignment off stack */
	skip_attribs(html_doc, file_index);
	break;

      case SAMP:
      case TELETYPE:
      case CODE:
      case KBD:
	
	debug (debug_parser, "parser: Found <KBD>");
	current_pattributes.monospace = TRUE;
	skip_attribs(html_doc, file_index);
	break;
	
      case SAMP|CLOSETAG:
      case TELETYPE|CLOSETAG:
      case CODE|CLOSETAG:
      case KBD|CLOSETAG:
	
	debug (debug_parser, "parser: Found </KBD>");
	current_pattributes.monospace = FALSE;
	skip_attribs(html_doc, file_index);
	break;

      case CITE:
      case VAR:
      case ITALIC:
      case EMPHASIS:

	debug (debug_parser, "parser: Found <I>");
	current_pattributes.emphasis = TRUE;
	skip_attribs(html_doc, file_index);
	break;

      case CITE|CLOSETAG:
      case VAR|CLOSETAG:
      case ITALIC|CLOSETAG:
      case EMPHASIS|CLOSETAG:

	debug (debug_parser, "parser: Found </I>");
	current_pattributes.emphasis = FALSE;
	skip_attribs(html_doc, file_index);
	break;

      case UNDERLINE:

	debug (debug_parser, "parser: Found <U>");
	current_pattributes.underline = TRUE;
	skip_attribs(html_doc, file_index);
	break;

      case UNDERLINE|CLOSETAG:

	debug (debug_parser, "parser: Found </U>");
	current_pattributes.underline = FALSE;
	skip_attribs(html_doc, file_index);
	break;

      case SUBSCRIPT:

	debug (debug_parser, "parser: Found <SUB>");
	current_pattributes.subscript = TRUE;
	skip_attribs (html_doc, file_index);
	break;

      case SUBSCRIPT|CLOSETAG:

	debug (debug_parser, "parser: Found </SUB>");
	current_pattributes.subscript = FALSE;
	skip_attribs(html_doc, file_index);
	break;

      case SUPSCRIPT:
 
	debug (debug_parser, "parser: Found <SUP>");
	current_pattributes.subscript = TRUE;
	skip_attribs (html_doc, file_index);
	break;

      case SUPSCRIPT|CLOSETAG:

	debug (debug_parser, "parser: Found </SUP>");
	current_pattributes.superscript = FALSE;
	skip_attribs(html_doc, file_index);
	break;

      case DL:

	debug (debug_parser, "parser: Found <DL>");
	push_list (&current_lattributes);
	current_lattributes.list_indent_level++;
	dl_top_indent_level = current_lattributes.list_indent_level;
	current_lattributes.item_number = 0;
	current_lattributes.dl_list_element = TRUE;
	current_lattributes.ul_list_element = FALSE;
	current_lattributes.ol_list_element = FALSE;

	if (neve_list)
	{
	  neve_in_chains *node = neve_list;
	  
	  while (node->previous && (node->type == end_link || node->type == start_list_element))
	    node = node->previous;
	  
	  debug (debug_parser, "parser.c: Trying to add PB for start of list, pointing at ...");
	  out_neve_type (node);
	  if (!node->previous)
	    new_neve_node(paragraph_break, &current_pattributes, &current_lattributes, current_href); /* YARK! Is this right ? */
	  else if (node->type == line_break)
	    node->type = paragraph_break;
	  else if (node->type != paragraph_break)
	    {
	      if (node->type == text)
		{
		  if (node->buf[0] == ' ' && node->buf[1] == '\0')
		    {
		      free (node->buf);
		      node->type = paragraph_break;
		    }
		  else
		    new_neve_node(paragraph_break, &current_pattributes, &current_lattributes, current_href);
		}
	      else
		new_neve_node(paragraph_break, &current_pattributes, &current_lattributes, current_href);
	    }
	}
	debug (debug_parser, "parser: Indent =%d, item_number=%d",
	       current_lattributes.list_indent_level, current_lattributes.item_number);
	parse_attribs (window, html_doc, file_index, neve_list, objfound, 
		       &current_basefont_size, &current_pattributes, &current_lattributes, current_href);
	/* Skip whitespace after UL but before LI. This tries to cope 
	   with people who put text between UL and the first LI */
	while (is_whitespace (html_doc[*file_index]))
	  (*file_index)++;
	
	if (!strncasecmp (html_doc+(*file_index), "<dt", 3) == 0)
	  current_lattributes.item_number++;

	break;


      case DT:

	debug (debug_parser, "parser: Found <DT>");
	
	if (neve_list)
	{
	  neve_in_chains *node = neve_list;
	  
	  debug (debug_parser, "parser: trying to add line_break for start_list_element");
	  while (node->previous && 
		 (node->type == end_link || node->type == start_list_element) && 
		 node->logical_attributes.item_number != 0)
	    node = node->previous;
	  debug (debug_parser, "parser: pointing at ...");
	  out_neve_type (node);
	  if (node->logical_attributes.item_number != 0) /* couldn't this be further out ie before the if(neve_list) check ? */
	    {
	      if (node->type == paragraph_break)
		node->type = line_break; /*YARK! Is this right? */
	      else if (node->type != line_break)
		{
		  if (node->type == text)
		    {
		      if (node->buf[0] == ' ' && node->buf[1] == '\0')
			{
			  free (node->buf);
			  node->type = line_break;
			}
		      else
			new_neve_node(line_break, &current_pattributes, &current_lattributes, current_href);
		    }
		  else
		    new_neve_node(line_break, &current_pattributes, &current_lattributes, current_href);
		}
	    }
	}
	
	current_lattributes.item_number++;
	current_lattributes.list_indent_level = dl_top_indent_level;
	new_neve_node(start_list_element, &current_pattributes, &current_lattributes, current_href);
	parse_attribs (window, html_doc, file_index, neve_list, objfound, 
		       &current_basefont_size, &current_pattributes, &current_lattributes, current_href);
	break;
	
	
      case DD:

	debug (debug_parser, "parser: Found <DD>");

	if (neve_list)
	  {
	    neve_in_chains *node = neve_list;
	    
	    debug (debug_parser, "parser: trying to add line_break for start_list_element");
	    while (node->previous && 
		   (node->type == end_link || node->type == start_list_element) && 
		   node->logical_attributes.item_number != 0)
	      node = node->previous;
	    debug (debug_parser, "parser: pointing at ...");
	    out_neve_type (node);
	    if (node->logical_attributes.item_number != 0)
	      {
		if (node->type == paragraph_break)
		  node->type = line_break;
		else if (node->type != line_break)
		  {
		    if (node->type == text)
		      {
			if (node->buf[0] == ' ' && node->buf[1] == '\0')
			  {
			    free (node->buf);
			    node->type = line_break;
			  }
			else
			  new_neve_node(line_break, &current_pattributes, &current_lattributes, current_href);
		      }
		    else
		      new_neve_node(line_break, &current_pattributes, &current_lattributes, current_href);
		  }
	      }
	  }
	current_lattributes.item_number++;
	current_lattributes.list_indent_level = dl_top_indent_level + 1;
	new_neve_node(start_list_element, &current_pattributes, &current_lattributes, current_href);
	parse_attribs (window, html_doc, file_index, neve_list, objfound, 
		       &current_basefont_size, &current_pattributes, &current_lattributes, current_href);
	break;
	
	
      case DL|CLOSETAG:

 	debug (debug_parser, "parser: Found </DL>");
	pop_list(&current_lattributes);
	if (current_lattributes.list_indent_level)
	  current_lattributes.item_number++;
	skip_attribs(html_doc, file_index);
	if (!current_lattributes.list_indent_level)
	  add_lbreaks(neve_list);
	break;

	
      case DIR:
      case MENU:
      case ULIST:

	debug (debug_parser, "parser: Found <UL>");
	push_list(&current_lattributes);
	current_lattributes.list_indent_level++;
	current_lattributes.item_number = 0;
	current_lattributes.ul_list_element = TRUE;
	current_lattributes.ol_list_element = FALSE;
	current_lattributes.dl_list_element = FALSE;

	debug (debug_parser, "parser: Checking back through neve to see if we should add a PB");
	if (neve_list)
	  {
	    neve_in_chains *node = neve_list;
	    while (node->previous && (node->type == end_link || node->type == start_list_element))
	      node = node->previous;
	    debug (debug_parser, "parser.c: Trying to add PB for start of list, pointing at ...");
	    out_neve_type (node);
	    if (!node->previous)
	      new_neve_node(paragraph_break, &current_pattributes, &current_lattributes, current_href);
	    else if (node->type == line_break)
	      node->type = paragraph_break;
	    else if (node->type != paragraph_break)
	      {
		if (node->type == text)
		  {
		    if (node->buf[0] == ' ' && node->buf[1] == '\0')
		      {
			free (node->buf);
			node->type = paragraph_break;
		      }
		    else
		      new_neve_node(paragraph_break, &current_pattributes, &current_lattributes, current_href);
		  }
		else
		  new_neve_node(paragraph_break, &current_pattributes, &current_lattributes, current_href);
	      }
	  }
 	debug (debug_parser, "parser: Indent=%d, item_number=%d",
	       current_lattributes.list_indent_level, current_lattributes.item_number);
	parse_attribs (window, html_doc, file_index, neve_list, objfound, 
		       &current_basefont_size, &current_pattributes, &current_lattributes, current_href);
        /* Skip whitespace after UL but before LI. This tries to cope 
	   with people who put text between UL and the first LI */
	while (is_whitespace (html_doc[*file_index]))
	  (*file_index)++;

	if (!strncasecmp (html_doc+(*file_index), "<li", 3) == 0)
	  current_lattributes.item_number++;
	
	break;
	
      case OLIST:

	debug (debug_parser, "parser: Found <OL>");
	push_list(&current_lattributes);
	current_lattributes.list_indent_level++;
	current_lattributes.item_number = 0;
	current_lattributes.ol_list_element = TRUE;
	current_lattributes.ul_list_element = FALSE;
	current_lattributes.dl_list_element = FALSE;
	
	if (neve_list)
	  {
	    neve_in_chains *node = neve_list;
	    
	    while (node->previous && (node->type == end_link || node->type == start_list_element))
	      node = node->previous;
	    debug (debug_parser, "parser.c: Trying to add PB for start of list, pointing at ...");
	    out_neve_type (node);
	    if (node->type == line_break)
	      {
		debug (debug_parser, "parser.c: Altering LB to PB.");
		node->type = paragraph_break;
	      }
	    else if (node->type != paragraph_break)
	      {
		if (node->type == text)
		  {
		    if (node->buf[0] == ' ' && node->buf[1] == '\0')
		      {
			free (node->buf);
			node->type = paragraph_break;
		      }
		    else
		      new_neve_node(paragraph_break, &current_pattributes, &current_lattributes, current_href);
		  }
		else
		  new_neve_node(paragraph_break, &current_pattributes, &current_lattributes, current_href);
	      }
	  }
	debug (debug_parser, "parser: Indent=%d, item_number=%d",
	       current_lattributes.list_indent_level, current_lattributes.item_number);
	parse_attribs (window, html_doc, file_index, neve_list, objfound, 
		       &current_basefont_size, &current_pattributes, &current_lattributes, current_href);
	/* Skip whitespace after UL but before LI. This tries to cope 
	   with people who put text between UL and the first LI */
	while (is_whitespace (html_doc[*file_index]))
	  (*file_index)++;
	
	if (!strncasecmp (html_doc+(*file_index), "<li", 3) == 0)
	  current_lattributes.item_number++;
	
	break;
	
      case DIV|CLOSETAG:
      case MENU|CLOSETAG:
      case ULIST|CLOSETAG:
      case OLIST|CLOSETAG:

 	debug (debug_parser, "parser: Found %s", (objfound == (ULIST|CLOSETAG)) ? "/UL" : "/OL");
	pop_list(&current_lattributes);
	if (current_lattributes.list_indent_level)
	  current_lattributes.item_number++;
	skip_attribs(html_doc, file_index);
	if (!current_lattributes.list_indent_level)
	  add_pbreaks(neve_list);
	else
	  add_lbreaks(neve_list);
	break;

      case LIST:

	debug (debug_parser, "parser: Found <LI>");
       
	if (neve_list)
	  {
	    neve_in_chains *node = neve_list;
	    
	    debug (debug_parser, "parser: trying to add line_break for start_list_element");
	    while (node->previous && 
		   (node->type == end_link || node->type == start_list_element) && 
		   node->logical_attributes.item_number != 0)
	      node = node->previous;
	    debug (debug_parser, "parser: pointing at ...");
	    out_neve_type (node);
	    if (node->logical_attributes.item_number != 0)
	      {
		if (node->type == paragraph_break)
		  node->type = line_break;
		else if (node->type != line_break)
		  {
		    if (node->type == text)
		      {
			if (node->buf[0] == ' ' && node->buf[1] == '\0')
			  {
			    free (node->buf);
			    node->type = line_break;
			  }
			else
			  new_neve_node(line_break, &current_pattributes, &current_lattributes, current_href);
		      }
		    else
		      new_neve_node(line_break, &current_pattributes, &current_lattributes, current_href);
		  }
	      }
	  }
	current_lattributes.item_number++;
	new_neve_node(start_list_element, &current_pattributes, &current_lattributes, current_href);
	parse_attribs (window, html_doc, file_index, neve_list, objfound, 
		       &current_basefont_size, &current_pattributes, &current_lattributes, current_href);
	break;
	
	
      case ANCHOR:

	debug (debug_parser, "parser: Found <A>");
	current_lattributes.link = TRUE;
	parse_attribs (window, html_doc, file_index, neve_list, objfound, 
		       &current_basefont_size, &current_pattributes, &current_lattributes, current_href);
	debug (debug_parser, "anchor: chref is %s", current_href);
	if (current_href[0]=='\0')
	  {
	    debug (debug_parser, "parser: No href found, link status := false");
	    current_lattributes.link = FALSE;
	    current_href[0] = '\0';
	  }
	break;

	
      case ANCHOR|CLOSETAG:

	debug (debug_parser, "parser: Found </A>");
	current_lattributes.link = FALSE;
	current_href[0]='\0';
	new_neve_node(end_link, &current_pattributes, &current_lattributes, current_href); /* Makes rendering easier */
	skip_attribs (html_doc, file_index);
	break;


      case PARAGRAPH: 

	debug (debug_parser, "parser: Found <P>");
	add_pbreaks(neve_list);
	pop_align (&current_pattributes, PARAGRAPH); /* Close the previous <P> */
	push_align (current_pattributes.alignment, PARAGRAPH); /* Push current alignment onto stack */
	parse_attribs (window, html_doc, file_index, neve_list, objfound, 
		       &current_basefont_size, &current_pattributes, &current_lattributes, current_href);
	break;


      case PARAGRAPH|CLOSETAG:

	debug (debug_parser, "parser: Found </P>");
	add_pbreaks(neve_list);
	pop_align (&current_pattributes, PARAGRAPH); /* Pop alignment off stack */
	skip_attribs (html_doc, file_index);
	break;


      case IMAGE:

	debug (debug_parser, "parser: Found <IMG>");
	new_neve_node(image, &current_pattributes, &current_lattributes, current_href);
	neve_list->physical_attributes.alignment = 0;     /* Ensure that TEXT/P/DIV align values are 
							     not carried into the image node */
	parse_attribs (window, html_doc, file_index, neve_list, objfound, 
		       &current_basefont_size, &current_pattributes, &current_lattributes, current_href);
	/* We'd like to put something as ALT text if it isn't given by the doc itself
	   Decided on {IMAGE:filename} */
	if (!neve_list->buf)
	  fill_in_alt(neve_list);
	break;


      case IMAGE|CLOSETAG: /* YARK can this be handled by the fallthrough last case ? */

	debug (debug_parser, "parser: Found </IMG>, ignoring!");
	skip_attribs (html_doc, file_index);  /* Not allowed, but just to be on the safe side */
	break;

	
      case BASEFONT:
	
	debug (debug_parser, "parser: Found <BASEFONT>");
	parse_attribs (window, html_doc, file_index, neve_list, objfound, 
		       &current_basefont_size, &current_pattributes, &current_lattributes, current_href);
	break;

	
      case FONT:
	
	debug (debug_parser, "parser: Found <FONT>");
	parse_attribs (window, html_doc, file_index, neve_list, objfound, 
		       &current_basefont_size, &current_pattributes, &current_lattributes, current_href);
	break;

	
      case FONT|CLOSETAG:
	
	debug (debug_parser, "parser: Found </FONT>");
	pop_font (&current_pattributes); /*YARK should pop into currents not the node */
	skip_attribs (html_doc, file_index);
	break;
	
	/* This is very very bad, but it is also the most efficient way of doing things */
	/* Do NOT, and i repeat NOT reorder or insert any additional cases until after the H1 */
      case H6:
	hlevel++;
      case H5:
	hlevel++;
      case H4:
	hlevel++;
      case H3:
	hlevel++;
      case H2:
	hlevel++;
      case H1:
	hlevel++;

	debug (debug_parser, "parser: Found a <H%d>", hlevel);
	hlevel = 14-hlevel; /* Scale from 1-6 to 8-13 */
	add_pbreaks(neve_list);
	push_header(current_pattributes.alignment, current_pattributes.font_level, current_pattributes.strong);
	current_pattributes.font_level = hlevel;
	if (objfound < H4)
	  current_pattributes.strong = TRUE;
	parse_attribs (window, html_doc, file_index, neve_list, objfound, 
		       &current_basefont_size, &current_pattributes, &current_lattributes, current_href);
	break;
	/* End evil code */

	
      case H1|CLOSETAG:
      case H2|CLOSETAG:
      case H3|CLOSETAG:
      case H4|CLOSETAG:
      case H5|CLOSETAG:
      case H6|CLOSETAG:
	
	debug (debug_parser, "parser: Found a close header tag");
	add_lbreaks(neve_list);
	pop_header(&current_pattributes);
	skip_attribs (html_doc, file_index);
	break;

	
      case HRULE:

	debug (debug_parser, "parser: Found a <HR>");
	add_lbreaks (neve_list);
	parse_attribs (window, html_doc, file_index, neve_list, objfound, 
		       &current_basefont_size, &current_pattributes, &current_lattributes, current_href);
	new_neve_node (horizontal_rule, &current_pattributes, &current_lattributes, current_href);
	break; /* YARK should be followed by carriage return */
	

      case BR:
	
	debug (debug_parser, "parser: Found a <BR>");
	new_neve_node(line_break, &current_pattributes, &current_lattributes, current_href);
	parse_attribs (window, html_doc, file_index, neve_list, objfound, 
		       &current_basefont_size, &current_pattributes, &current_lattributes, current_href);
	break;
	
      case TITLE:

	debug (debug_parser, "parser: Found a <TITLE>");
	title_flag = TRUE;
	skip_attribs (html_doc, file_index);
	break;

      case TITLE|CLOSETAG:

	debug (debug_parser, "parser: Found a </TITLE>");
	title_flag = FALSE;
	skip_attribs (html_doc, file_index);
	break;

      case BLOCKQUOTE:

	debug (debug_parser, "parser: Found a <BLOCKQUOTE>");
	current_pattributes.blockquote = TRUE;
	skip_attribs (html_doc, file_index);
	break;

      case BIG:

	debug (debug_parser, "parser: Found a <BIG>");
	push_header (current_pattributes.alignment, current_pattributes.font_level, current_pattributes.strong);
	if (current_pattributes.font_level < 8 )
	  { /* We're in normal text, not a header */
	    current_pattributes.font_level ++;
	    if (current_pattributes.font_level == 8)
	      current_pattributes.font_level = 7;
	  }
	else
	  { /* We're in a header */
	    current_pattributes.font_level--;
	    if (current_pattributes.font_level == 7)
	      current_pattributes.font_level = 8;
	  }
	skip_attribs (html_doc, file_index);
	break;

      case SMALL:

	debug (debug_parser, "parser: Found a <SMALL>");
	push_header (current_pattributes.alignment, current_pattributes.font_level, current_pattributes.strong);
	if (current_pattributes.font_level < 8 )
	  {
	    current_pattributes.font_level--;
	    if (current_pattributes.font_level == 0)
	      current_pattributes.font_level = 1;
	  }
	else
	  { current_pattributes.font_level++;
	  if (current_pattributes.font_level == 14)
	    current_pattributes.font_level = 13;
	  }
	skip_attribs (html_doc, file_index);
	break;
	
      case BIG|CLOSETAG:
      case SMALL|CLOSETAG:
	
	debug (debug_parser, "parser: Found a </BIG>");
	pop_header (&current_pattributes);
	skip_attribs (html_doc, file_index);
	break;

      case BLOCKQUOTE|CLOSETAG:
	
	debug (debug_parser, "parser: Found </BLOCKQUOTE>");
	current_pattributes.blockquote = FALSE;
	skip_attribs(html_doc, file_index);
	break;

      case ADDRESS:

	debug (debug_parser, "parser: Found <ADDRESS>");
	current_pattributes.emphasis = TRUE;
	skip_attribs(html_doc, file_index);
	break;
	
      case ADDRESS|CLOSETAG:

	debug (debug_parser, "parser: Found </ADDRESS>");
	current_pattributes.emphasis = FALSE;
	add_pbreaks(neve_list);
	skip_attribs(html_doc, file_index);
	break;
	
      case COMMENT:
	break;

      case BODY:
	debug (debug_parser, "parser: Found <BODY>");
	parse_attribs (window, html_doc, file_index, neve_list, objfound, 
		       &current_basefont_size, &current_pattributes, &current_lattributes, current_href);
	break;
	
      case SCRIPT:
	debug (debug_parser, "parser: Found <SCRIPT>");
	skip_attribs (html_doc, file_index);
	
	while (((objfound=scan(html_doc, file_index, text_buffer, current_lattributes.preformatted, previous_tag)) != END_OF_DOC) && 
	       (objfound != (SCRIPT|CLOSETAG)))
	  if (objfound < COMMENT)
	    skip_attribs (html_doc, file_index);
	  
	if (objfound == (SCRIPT|CLOSETAG))
	  debug (debug_parser, "parser: Found </SCRIPT");
	  skip_attribs (html_doc, file_index);
	break;
	    

      default:
	debug (debug_parser, "parser: something we don't deal with DOH!");
	skip_attribs (html_doc, file_index);
	break;
      }

    previous_tag = objfound;

  }    
  return 0;
}



int parse_html (char *html_doc, BrowserWindow *paula)
{
  /* This is the entry point that initialises all our variables for a
   * new document */

  char *text_buffer;
  long file_index = 0;

  neve_head = NULL; 
  neve_list = NULL;
  text_buffer = xmalloc(BUF_SIZE);
  current_href = xmalloc(BUF_SIZE); /* How big should this be */

  font_stack = NULL;
  header_stack = NULL;
  list_stack = NULL;
  align_stack = NULL;

  /* Set up the Logical attributes to start with */
  current_lattributes.link = FALSE;
  current_lattributes.link_visited = FALSE;
  current_lattributes.ul_list_element = FALSE;
  current_lattributes.ol_list_element = FALSE;
  current_lattributes.dl_list_element = FALSE;
  current_lattributes.list_indent_level = 0;
  /* Set up the Physical attributes */
  current_pattributes.font_level = DEFAULT_FONT_LEVEL;
  current_pattributes.strong = FALSE;
  current_pattributes.emphasis = FALSE;
  current_pattributes.code = FALSE;
  current_pattributes.alignment = left;
  current_pattributes.underline = FALSE;
  current_pattributes.subscript = FALSE;
  current_pattributes.superscript = FALSE;
  current_pattributes.strike = FALSE;
  current_pattributes.draw_colour.red = 0;
  current_pattributes.draw_colour.green = 0;
  current_pattributes.draw_colour.blue = 0;
  current_pattributes.blockquote = FALSE;

  expect_document(paula, html_doc, &file_index, text_buffer);
  free (text_buffer);
  debug ( debug_parser, "Frankly ... WE ROCK!" );
  output_list (neve_head);
  debug (debug_parser, 
	  "parser: Succesfully finished parsing, passing struct to rendered" ) ;
  lisa_take_neve (neve_head, paula);
  
  return 0;
}
