/* YAWN, Yet Another Web Navigator */
/* Copyright (C) 1998, James Troup <jjtroup@comp.brad.ac.uk>
 *                     Lee Willis <lwillis@comp.brad.ac.uk> */

/* This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include <stdio.h>
#include <string.h>
#include <gtk/gtkwidget.h>
#include "utils.h"
#include "neve.h"
#include "lisa.h"
#include "render.h"

/* Declarations for internal functions */

void render_alignment (const neve_alignment alignment, const GtkLisa * lisa, const int current_x,
		       const int y_coordinate, const int nicole_width, const int font_height,
		       const int line_space);
int render_neve (const louise * line_cache, neve_in_chains * neve, const boolean first_neve,
		 GtkLisa * lisa, const GdkRectangle * area, const int nicole_width,
		 int *current_x, const int real_y, neve_alignment * alignment);
int render_text (const louise * line_cache, neve_in_chains * neve, const boolean first_neve,
		 GtkLisa * lisa, const GdkRectangle * area, const int nicole_width,
		 int *current_x, const int real_y, neve_alignment * alignment);
void set_colour (GtkLisa * lisa, const neve_in_chains *neve, const boolean in_link);
void render_bullet (GtkLisa * lisa, const neve_in_chains * neve, int x_coordinate, int y_coordinate);

/* Render a single line of text. This code is hairy */

void
render_line (louise * line_cache, GtkLisa * lisa, GdkRectangle * area, int *real_y)
{

  boolean first_neve = TRUE;
  neve_in_chains *neve;
  neve_alignment alignment = left;
  int i, line_space, current_x, indent_amount, nicole_width;

  gdk_window_get_size (lisa->nicole, &nicole_width, NULL);
  nicole_width -= 4;
  neve = line_cache->start_of_line;
  line_space = 2;
  indent_amount = 0;

  current_x = 0;
  debug (debug_render, "render_line: starting (nicole_width is %d, y is %d or %d)", 
	 nicole_width, *real_y, line_cache->y_coordinate);
  
  do
    {
      debug (debug_render, "render_line: loops begins");
      i = render_neve (line_cache, neve, first_neve, lisa, area, nicole_width, &current_x,
		       *real_y + lisa->offset, &alignment);
      first_neve = FALSE;
      if (neve->next && i > 0)
	{
	  debug (debug_render, "render_line: moving onto next node");
	  neve = neve->next;
	}
      else
	i = 0;
    }
  while (i > 0);

  if (alignment != left)
    {
      debug (debug_render, "nicole width is %d", nicole_width);
      render_alignment (alignment, lisa, current_x, *real_y + lisa->offset + 2, nicole_width, line_cache->max_font_height, 
			line_space);
    }

  debug (debug_render, "real y is %d, adding %d + %d (%d), becomes %d", 	
	 *real_y, line_cache->max_font_height, line_space, line_cache->max_font_height + line_space, 
  /* 	 ((*real_y) +*/ line_cache->max_font_height + line_space + (*real_y)); 
  
  area->y += line_cache->max_font_height + line_space;
  (*real_y) += line_cache->max_font_height + line_space;


}

/* Handle alignment. 
 *
 * We handle text aligment by blitting already-rendered text around. 
 *
 * No support for a) image alignment (scary issue) or b) justify. */

void
render_alignment (const neve_alignment alignment, const GtkLisa * lisa, const int current_x,
		  const int y_coordinate, const int nicole_width, const int font_height,
		  const int line_space)
{

  int width_of_line = current_x;
  int height_of_line = font_height + line_space;
  int new_x;
  
  if (current_x > nicole_width)
    {
      /* FIXME: we're rendering beyond nicole width */
      error ("render_alignment: Abuse James; current_x -> nicole_width (%d > %d)",
	     current_x, nicole_width);
      error ("                  if two men say they're Jesus, one of them must be wrong.");
      current_x = nicole_width;
    }

  if (alignment == centre)
    new_x = (nicole_width - current_x) / 2;
  else if (alignment == right)
    {
      new_x = nicole_width - current_x;
      if (new_x < 0)
	new_x = 0;
    }
  else if (alignment == justify)
    /* Haha, I *so* do not think so.  Besides this was only in 3.0
     * and dropped for 3.2+ and is arguably not something which
     * should be in HTML (being handled, instead, by CSS1/DSSSL and
     * friends). */
    return;
  else
    /* Anything else is either bogus, or one of the image alignments
     * which we don't yet handle. */
    return;

  debug (debug_render, "render_alignment: nicole width = %d, current x = %d, new x = %d, current y = %d, font height = %d, line space =%d",
	 nicole_width, current_x, new_x, y_coordinate, font_height, line_space);
  debug (debug_render, "render_alignment: blitting from %dx%d -> %dx%d with %d width and %d height",
	 0, y_coordinate, new_x, y_coordinate, width_of_line, height_of_line);
  gdk_draw_pixmap (lisa->nicole, lisa->gc,
		   lisa->nicole, 0, y_coordinate,
		   new_x, y_coordinate, width_of_line, height_of_line);
  debug (debug_render, "render_alignment: clearing from %dx%d with %d width and %d height",
	 0, y_coordinate, new_x, height_of_line);
  if (new_x > 0 && height_of_line > 0)
    gdk_window_clear_area (lisa->nicole,
			   0, y_coordinate,
			   new_x, height_of_line);
}



/* This procedure is called by render_line() for each neve node.  It
 * simply switches on the type of the neve node and calls functions
 * accordingly. It could be louise_parse_neve() except for differences
 * in the hairy arguments we're passing down */

int
render_neve (const louise * line_cache, neve_in_chains * neve, const boolean first_neve,
	     GtkLisa * lisa, const GdkRectangle * area, const int nicole_width,
	     int *current_x, const int real_y, neve_alignment * alignment)
{

  int return_value;

  if (first_neve && neve->logical_attributes.list_indent_level > 0)
    (*current_x) += 25 + (neve->logical_attributes.list_indent_level * 10);

  switch (neve->type)
    {
    case text:
      debug (debug_render, "render_neve:  text found.");
      return_value = render_text (line_cache, neve, first_neve, lisa, area,
			 nicole_width, &(*current_x), real_y, &(*alignment));
      break;
    case image:
      debug (debug_render, "render_neve: image found.");
      return_value = render_text (line_cache, neve, first_neve, lisa, area,
			 nicole_width, &(*current_x), real_y, &(*alignment));
      break;
    case horizontal_rule:
      debug (debug_render, "render_neve: horizontal rule found.");
      debug (debug_render, "render_neve: horizontal rule rendered: %dx%d+%d.%d.", 
	     *current_x, real_y + 6,
	     nicole_width, real_y + 6);
      gdk_draw_line (lisa->nicole, lisa->gc,
		     *current_x, real_y + 6,
		     nicole_width, real_y + 6);
      return -1;
      break;
    case line_break:
      debug (debug_render, "render_neve: line break return found.");
      return -1;
      break;
    case paragraph_break:
      debug (debug_render, "render_neve: paragraph break return found.");
      return -1;
      break;
    case end_link:
      debug (debug_render, "render_neve: end link return found.");
      set_colour (lisa, neve, FALSE);
      return 1;
      break;
    case start_list_element:
      debug (debug_render, "render_neve: start_list_element found.");
      render_bullet (lisa, neve, *current_x, real_y);
      return exhausted_node;
      break;
    default:
      fubar (NONSYS, "render_neve: found %d, which I clearly don't recognise. outie :-(",
	     neve->type);
      return 1;
    }

  return return_value;

}

/* Generic text handling.  There is a *lot* of code duplication with
 * the neve2louise code, since this is doing everything the
 * neve2louise code does, and then draw it.  It would _really_ be a
 * good plan to merge these two, and have the actual rendering
 * conditional on an argument.
 *
 * Also the number of arguments is plainly silly, and could do with
 * some rationalization, somehow. */

int
render_text (const louise * line_cache, neve_in_chains * neve, const boolean first_neve,
	     GtkLisa * lisa, const GdkRectangle * area, const int nicole_width,
	     int *current_x, const int real_y, neve_alignment * alignment)
{

  char buf[1000];		/* FIXME: this is tacky */
  char fontname[256];
  GdkFont *font;
  int font_size;
  int start_of_region, end_of_region, x_coordinate, old_x;
  int return_value = 1;
  int max_width;
  int y_offset = 0; /* To compensate for changes in font size, mid-line */
  x_coordinate = *current_x;
  old_x = x_coordinate;

  /* Check alignment */

  debug (debug_render, "render_text: alignment = %d, neve's alignment = %d (header level = %d)", 
	 *alignment, neve->physical_attributes.alignment, neve->physical_attributes.font_level);
  if (neve->physical_attributes.alignment != *alignment)
    {
      *alignment = neve->physical_attributes.alignment;
      debug (debug_render, "render_text: alignment => %d", *alignment);
    }
  debug (debug_render, "x_coordinate is %d", x_coordinate);

  /* Check for a link */

  if (neve->logical_attributes.link)
    {
      debug (debug_render, "render_text: is a link, rendering at %d", *current_x);
      set_colour (lisa, neve, TRUE);
    }
  else
    {
      set_colour (lisa, neve, FALSE);
    }

  if (area->width > nicole_width)
    /* FIXME: why does this happen? */
    max_width = nicole_width;
  else if (area->width == nicole_width || area->width > nicole_width)
    max_width = nicole_width;
  else
    max_width = area->width;

  if (first_neve)
    {
      debug (debug_render, "render_text: first neve found, setting up appropriate variables");
      start_of_region = line_cache->offset_into_first_neve;
      end_of_region = line_cache->offset_into_first_neve;
      debug (debug_render, "render_text: start of liner is %d, end of line is %d", start_of_region, end_of_region);
    }
  else
    {
      debug (debug_render, "render_text: this is the second or later neve, i & h => 0");
      start_of_region = 0;
      end_of_region = 0;
    }

  debug (debug_render, "render_text: font level is %d", neve->physical_attributes.font_level);
  font_size = font_levels[neve->physical_attributes.font_level - 1];
  debug (debug_render, "render_text: font size --> %d", font_size);

  sprintf (fontname, "-adobe-%s-%s-%s-normal--%d-100-*-*-*-*-*-*",
	   neve->physical_attributes.monospace ? "courier" : "helvetica", 
	   neve->physical_attributes.strong ? "bold" : "medium",
	   neve->physical_attributes.emphasis ? "o" : "r",
	   font_size);

  debug (debug_render, "render_text: fontname is \"%s\"\n", fontname);

  font = gdk_font_load (fontname);

  if (font == NULL)
    {
      fubar (NONSYS, "render_text: couldn't load font. outie");
    }

  debug (debug_render, "render_text (linda[2]): %d => %d + %d", font_size, font->ascent, font->descent);

  if ((font->ascent + font->descent) < line_cache->max_font_height)
    y_offset = line_cache->max_font_height - (font->ascent + font->descent);

  /* There are three possible endings here:

   *  1) The text runs out.  (neve->buf[i] = '\0')
   *  2) We run out of area to expose in (x_coordinate > area->width)
   *  3) We run out of line to render on (x_coordinate > nicole_width)
   *
   * With (3), unlike the others where we just stop or go on
   * to the next neve, we have to break at whitespace and not
   * mid-word.
   *
   * With both (2) and (3) we need to skip back one, otherwise
   * we become out of sync with the line cache and get word
   * doubling.  I believe this is a bug in lisa_take_neve().
   *
   * FIXME
   */

  debug (debug_render, "foobarbazbat: area->width = %d", area->width);
  while (x_coordinate < max_width &&
	 neve->buf[end_of_region] != '\0')
    {
      x_coordinate += gdk_char_width (font, neve->buf[end_of_region]);
      end_of_region++;
    }

  if (neve->buf[end_of_region] != '\0')		/* Line is finished */
    {
      debug (debug_render, "render_text: I think the neve buffer is not finished.");
      return_value = 0;
    }
  else
    /* Exhausted this neve node */
    {
      debug (debug_render, "render_text: I think the neve buffer is finished.");
    }

  debug (debug_render, "^x_coordinate is %d", x_coordinate);
  if (x_coordinate >= nicole_width)
    backup_to_last_whitespace (neve, start_of_region, &end_of_region, &x_coordinate, font);
  debug (debug_render, "^x_coordinate is %d", x_coordinate);

  debug (debug_render, "FOOBAR: end_of_region is %d, start_of_region is %d", end_of_region, start_of_region);
  strncpy (buf, neve->buf + start_of_region, end_of_region - start_of_region);
  buf[(end_of_region - start_of_region)] = '\0';
  debug (debug_render, "**:%s:**", buf);
  debug (debug_render, "$$:%s:$$", neve->buf);
  debug (debug_render, "$Rendering at %d, %d", *current_x, real_y + font->ascent + y_offset);

  gdk_draw_string (lisa->nicole, font,
		   lisa->gc,
		   *current_x, real_y + font->ascent + y_offset, buf);
  debug (debug_render, "^X is %d", *current_x);

  (*current_x) += (x_coordinate - old_x);
  debug (debug_render, "^X is %d", *current_x);

  return return_value;

}

/* Render a bullet for an unordered list. */

void
render_bullet (GtkLisa * lisa, const neve_in_chains * neve, int x_coordinate, int y_coordinate)
{

  /* Bullets can appear in a link, but there's no way I'm writing the
   * code to make them clickable, so we won't render them in link
   * colour even if they are in a link. */

  set_colour (lisa, neve, FALSE);

  if (neve->logical_attributes.ul_list_element)
    {
      x_coordinate -= 10;

      if (neve->logical_attributes.list_indent_level == 1)
	{
	  debug (debug_render, "render_neve: drawing filled circle at %dx%d %d.%d",
		 x_coordinate, y_coordinate + 3, 7, 7);
	  gdk_draw_arc (lisa->nicole, lisa->gc, TRUE,
			x_coordinate, y_coordinate + 3, 7, 7, 0, 23040);	
	  /* 23040 = 360 * 64; don't ask me why, I have to multiply by
	   * 64, it's just something gdk_draw_arc()/XDrawArc wants done. */
	}
      else if (neve->logical_attributes.list_indent_level == 2)
	{
	  debug (debug_render, "render_neve: drawing circle at %dx%d %d.%d",
		 x_coordinate, y_coordinate + 4, 6, 6);
	  gdk_draw_arc (lisa->nicole, lisa->gc, FALSE,
			x_coordinate, y_coordinate + 4, 6, 6, 0, 23040);
	}
      else
	{
	  debug (debug_render, "render_neve: drawing box at %dx%d %d.%d",
		 x_coordinate, y_coordinate + 4, 7, 7);
	  gdk_draw_rectangle (lisa->nicole, lisa->gc, FALSE,
			      x_coordinate, y_coordinate + 4, 7, 7);
	}
    }
      
}

/* Skip back through the neve node to the first whitespace. */

void
backup_to_last_whitespace (const neve_in_chains * neve, const int start_of_region, int *end_of_region, int *x_coordinate, GdkFont * font)
{

  /* This will break if text is split over neve nodes; e.g.
   *
   * "Warren Ellis writes damn fiNE comics"
   * (where the "NE" of ``fiNE'' is in bold)
   *
   * If the geometry requires it, this will split across lines thusly:
   *
   * """Warren Ellis writes damn fi\n\
   *    NE comics"""
   *
   * This is sub-optimal, assuming the assumption that words shouldn't
   * be broken across lines holds. 
   *
   * This also fails most hideously to take into accounts the evilness
   * known as &nbsp;.
   */

  int last_whitespace = *end_of_region;

  while (last_whitespace >= start_of_region && neve->buf[last_whitespace] != ' ')
    {
      (*x_coordinate) -= gdk_char_width (font, neve->buf[last_whitespace]);
      last_whitespace--;
    }

  /* Iff we haven't hit the start of the neve node, end_region becomes
   * the point of the last whitespace */
  
  if (last_whitespace != (start_of_region - 1))
    *end_of_region = last_whitespace;

}

void
set_colour (GtkLisa * lisa, const neve_in_chains *neve, const boolean in_link)
{
  GdkColor foreground_colour;
  GdkColormap *cmap;

  cmap = gtk_widget_get_colormap (&lisa->widget);

  if (in_link)
    {
      if (!gdk_color_alloc (cmap, lisa->link_colour))
	fubar (NONSYS, "set_colour: couldn't allocate colour.  outie.");
      gdk_gc_set_foreground (lisa->gc, lisa->link_colour);
    }
  else
    {
      foreground_colour.red = neve->physical_attributes.draw_colour.red * 256;
      foreground_colour.green = neve->physical_attributes.draw_colour.green * 256;
      foreground_colour.blue = neve->physical_attributes.draw_colour.blue * 256;
      if (!gdk_color_alloc (cmap, &foreground_colour))
	fubar (NONSYS, "set_colour: couldn't allocate colour.  outie.");
      gdk_gc_set_foreground (lisa->gc, &foreground_colour);
    }
}
