/* YAWN, Yet Another Web Navigator */
/* Copyright (C) 1998, James Troup <jjtroup@comp.brad.ac.uk>
 *                     Lee Willis <lwillis@comp.brad.ac.uk> */

/* This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include <stdio.h>
#include <stdarg.h>
#include <string.h>
#include <errno.h>
#include "globals.h"
#include "utils.h"

/*****************************************************************************
 * function  : debug
 * arguments : debug level and {s,f,}printf style format string
 * results   : 
 * purpose   : output debugging information if the global debug level is
 *             high enough
 * caveats   : this function is always present, and always called,
 *             regardless of the debug level.
 ***************************************************************************** */

void
debug (const int which, const char *fmt,...)
{

  va_list args;

  if (!(which & debug_level))
    return;

  fprintf (stderr, "D0%05o: ", which);
  va_start (args, fmt);
  vfprintf (stderr, fmt, args);
  va_end (args);
  putc ('\n', stderr);
}

/*****************************************************************************
 * function  : debug_help
 * arguments : 
 * results   : 
 * purpose   : give information on debug levels
 * caveats   : 
 ***************************************************************************** */

void
debug_help (void)
{
  
  printf ("YAWN debugging option, --debug=<octal value> or -D<octal value>:\n\
  number  description\n\
       1  General warnings\n\
       2  Debug information for Neve handling\n\
       4  Debug information for the renderer\n\
      10  Debug information for Lisa\n\
      20  Debug information for Lisa's scrolling\n\
      40  Debug information for Lisa's expose code\n\
     100  Debug information for the HTML parser\n\
     200  Debug information for the HTML attribute parsing\n\
     400  Debug information for the HTML scanner\n\
    1000  Debug information for the network and file handling\n\
\n\
Combine debug options with bitwise or.\n\
Meanings and values subject to change at any time.\n");
  exit (0);
}

/*****************************************************************************
 * function  : error
 * arguments : {s,f,}printf style format string
 * results   : 
 * purpose   : output an error string
 * caveats   : does not die.  see fubar() if you have a fatal error.
 ***************************************************************************** */

void
error (const char *fmt,...)
{
  va_list args;

  fprintf (stderr, "[%s]: ", program_name);
  va_start (args, fmt);
  vfprintf (stderr, fmt, args);
  va_end (args);
  putc ('\n', stderr);
}

/*****************************************************************************
 * function  : error
 * arguments : system? flag and {s,f,}printf style format string
 * results   : 
 * purpose   : output an error string and die.  flag is used to indicate
 *             a system error.
 * caveats   : 
 ***************************************************************************** */

void
fubar (unsigned short flag, const char *fmt,...)
{

  unsigned short syserr = flag & SYSERR;
  va_list args;

  fprintf (stderr, "[%s]: ", program_name);
  va_start (args, fmt);
  vfprintf (stderr, fmt, args);
  va_end (args);
  if (syserr)
    fprintf (stderr, " : %s", strerror (errno));
  putc ('\n', stderr);

  exit (1);
}
