/* xmalloc.c -- malloc with out of memory checking
   Copyright (C) 1990, 91, 92, 93, 94, 95, 96 Free Software Foundation, Inc.
   Copyright (C) 1997, 1998, James Troup <jjtroup@comp.brad.ac.uk>
                             Lee Willis <lwillis@comp.brad.ac.uk> */

/* This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include <sys/types.h>
#include <stdlib.h>
#include "utils.h"


/*****************************************************************************
 * function  : fixup_null_alloc
 * arguments : pointer
 * results   : pointer to the memory
 * purpose   : try to recover if xmalloc recevied an error from malloc
 *             or we tried to xmalloc zero bytes
 * caveats   : if we can't allocate the memory requested , we just fall
 *             over and die.
 ***************************************************************************** */

void *
fixup_null_alloc (size_t n)
{
  void *p;

  p = 0;
  if (n == 0)
    p = malloc ((size_t) 1);
  if (p == 0)
    fubar (SYSERR, "Memory exhausted");
  return p;
}


/*****************************************************************************
 * function  : xmalloc
 * arguments : amount of memory to malloc
 * results   : pointer to the memory
 * purpose   : dynamically allocate X bytes of memory, *and* check for errors.
 * caveats   : see fixup_null_alloc()
 ***************************************************************************** */

void *
xmalloc (size_t n)
{
  void *p;

  p = malloc (n);
  if (p == 0)
    p = fixup_null_alloc (n);
  return p;
}
